# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines the core functionality of the dispatcher
"""

import asyncio
import traceback
from datetime import datetime, timezone
from typing import Dict, List, Tuple

import networkx as nx

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.defaults import WAIT_EDGE_NAME, parameter_prefix
from covalent._shared_files.util_classes import RESULT_STATUS

from . import data_manager as datasvc
from . import runner_ng
from .data_modules import graph as tg_utils
from .data_modules import job_manager as jbmgr
from .dispatcher_modules.caches import _pending_parents, _sorted_task_groups, _unresolved_tasks
from .runner_modules.cancel import cancel_tasks

app_log = logger.app_log
log_stack_info = logger.log_stack_info
_global_status_queue = None
_status_queues = {}
_futures = {}

_global_event_listener = None

SYNC_DISPATCHES = get_config("dispatcher.use_async_dispatcher") == "false"


# Domain: dispatcher
async def _get_abstract_task_inputs(dispatch_id: str, node_id: int, node_name: str) -> dict:
    """Return placeholders for the required inputs for a task execution.

    Args:
        dispatch_id: id of the current dispatch
        node_id: Node id of this task in the transport graph.
        node_name: Name of the node.

    Returns: inputs: Input dictionary to be passed to the task with
        `node_id` placeholders for args, kwargs. These are to be
        resolved to their values later.
    """

    abstract_task_input = {"args": [], "kwargs": {}}

    for edge in await tg_utils.get_incoming_edges(dispatch_id, node_id):
        parent = edge["source"]

        d = edge["attrs"]

        if d["edge_name"] != WAIT_EDGE_NAME:
            if d["param_type"] == "arg":
                abstract_task_input["args"].append((parent, d["arg_index"]))
            elif d["param_type"] == "kwarg":
                key = d["edge_name"]
                abstract_task_input["kwargs"][key] = parent

    sorted_args = sorted(abstract_task_input["args"], key=lambda x: x[1])
    abstract_task_input["args"] = [x[0] for x in sorted_args]

    return abstract_task_input


# Domain: dispatcher
async def _handle_completed_node(dispatch_id: str, node_id: int):
    next_task_groups = []
    app_log.debug(f"Node {node_id} completed")

    parent_gid = (await datasvc.electron.get(dispatch_id, node_id, ["task_group_id"]))[
        "task_group_id"
    ]
    for child in await tg_utils.get_node_successors(dispatch_id, node_id):
        node_id = child["node_id"]
        gid = child["task_group_id"]
        app_log.debug(f"dispatch {dispatch_id}: parent gid {parent_gid}, child gid {gid}")
        if parent_gid != gid:
            now_pending = await _pending_parents.decrement(dispatch_id, gid)
            if now_pending < 1:
                app_log.debug(f"Queuing task group {gid} for execution")
                next_task_groups.append(gid)

    return next_task_groups


# Domain: dispatcher
async def _handle_failed_node(dispatch_id: str, node_id: int):
    app_log.debug(f"Node {dispatch_id}:{node_id} failed")
    app_log.debug("8A: Failed node upsert statement (run_planned_workflow)")


# Domain: dispatcher
async def _handle_cancelled_node(dispatch_id: str, node_id: int):
    app_log.debug(f"Node {dispatch_id}:{node_id} cancelled")
    app_log.debug("9: Cancelled node upsert statement (run_planned_workflow)")


# Domain: dispatcher
async def _get_initial_tasks_and_deps(dispatch_id: str) -> Tuple[int, int, Dict]:
    """Compute the initial batch of tasks to submit and initialize each task's dep count

    Returns: (num_tasks, ready_nodes, pending_parents) where num_tasks is
        the total number of tasks in the graph, ready_nodes is the
        initial list of tasks to dispatch, and pending_parents is a map
        from `node_id` to the number of parents that have yet to
        complete.

    """

    # Number of pending predecessor nodes for each task group
    pending_parents = {}

    g_node_link = await tg_utils.get_nodes_links(dispatch_id)
    g = nx.readwrite.node_link_graph(g_node_link)

    # Topologically sort each task group
    sorted_task_groups = {}
    for node_id in nx.topological_sort(g):
        gid = g.nodes[node_id]["task_group_id"]
        if gid not in sorted_task_groups:
            sorted_task_groups[gid] = [node_id]
            pending_parents[gid] = 0
        else:
            sorted_task_groups[gid].append(node_id)

    for node_id in g.nodes:
        parent_gid = g.nodes[node_id]["task_group_id"]
        for succ, datadict in g.adj[node_id].items():
            child_gid = g.nodes[succ]["task_group_id"]

            if parent_gid != child_gid:
                n_edges = len(datadict.keys())
                pending_parents[child_gid] += n_edges

    initial_task_groups = [gid for gid, d in pending_parents.items() if d == 0]
    app_log.debug(f"Sorted task groups: {sorted_task_groups}")
    return initial_task_groups, pending_parents, sorted_task_groups


# Domain: dispatcher
async def _submit_task_group(dispatch_id: str, sorted_nodes: List[int], task_group_id: int):
    # Handle parameter nodes
    # Get name of the node for the current task
    node_name = (await datasvc.electron.get(dispatch_id, sorted_nodes[0], ["name"]))["name"]
    app_log.debug(f"7A: Node name: {node_name} (run_planned_workflow).")

    # Handle parameter nodes
    if node_name.startswith(parameter_prefix):
        if len(sorted_nodes) > 1:
            raise RuntimeError("Parameter nodes cannot be packed")

        app_log.debug("7C: Encountered parameter node {node_id}.")
        app_log.debug("8: Starting update node (run_planned_workflow).")

        ts = datetime.now(timezone.utc)
        node_result = {
            "node_id": sorted_nodes[0],
            "start_time": ts,
            "end_time": ts,
            "status": RESULT_STATUS.COMPLETED,
        }
        await datasvc.update_node_result(dispatch_id, node_result)
        app_log.debug("8A: Update node success (run_planned_workflow).")

    else:
        # Nodes whose values have already been resolved
        known_nodes = []

        # Skip the group if all task outputs can be reused from a
        # previous dispatch (for redispatch).
        statuses = await datasvc.electron.get_bulk(dispatch_id, sorted_nodes, ["status"])
        incomplete = list(
            filter(lambda record: record["status"] != RESULT_STATUS.PENDING_REUSE, statuses)
        )

        if incomplete:
            # Gather inputs for each task and send the task spec sequence to the runner
            task_specs = []

            sorted_nodes_set = set(sorted_nodes)

            for node_id in sorted_nodes:
                app_log.debug(f"Gathering inputs for task {node_id} (run_planned_workflow).")

                abs_task_input = await _get_abstract_task_inputs(dispatch_id, node_id, node_name)

                executor_attrs = await datasvc.electron.get(
                    dispatch_id,
                    node_id,
                    ["executor", "executor_data"],
                )
                selected_executor = executor_attrs["executor"]
                selected_executor_data = executor_attrs["executor_data"]
                task_spec = {
                    "function_id": node_id,
                    "name": node_name,
                    "args_ids": abs_task_input["args"],
                    "kwargs_ids": abs_task_input["kwargs"],
                }
                # Task inputs that don't belong to the task group have already beeen resolved
                external_task_args = filter(
                    lambda x: x not in sorted_nodes_set, abs_task_input["args"]
                )
                known_nodes.extend(external_task_args)
                external_task_kwargs = filter(
                    lambda x: x not in sorted_nodes_set, abs_task_input["kwargs"].values()
                )
                known_nodes.extend(external_task_kwargs)

                task_specs.append(task_spec)

            app_log.debug(
                f"Submitting task group {dispatch_id}:{task_group_id} ({len(sorted_nodes)} tasks) to runner"
            )
            app_log.debug(f"Using new runner for task group {task_group_id}")

            known_nodes = list(set(known_nodes))
            coro = runner_ng.run_abstract_task_group(
                dispatch_id=dispatch_id,
                task_group_id=task_group_id,
                task_seq=task_specs,
                known_nodes=known_nodes,
                selected_executor=[selected_executor, selected_executor_data],
            )

            asyncio.create_task(coro)
        else:
            ts = datetime.now(timezone.utc)
            for node_id in sorted_nodes:
                app_log.debug(f"Skipping already completed node {dispatch_id}:{node_id}")
                node_result = {
                    "node_id": node_id,
                    "start_time": ts,
                    "end_time": ts,
                    "status": RESULT_STATUS.COMPLETED,
                }
                await datasvc.update_node_result(dispatch_id, node_result)
                app_log.debug("8A: Update node success (run_planned_workflow).")


async def _plan_workflow(dispatch_id: str) -> None:
    """
    Function to plan a workflow according to a schedule.
    Planning means to decide which executors (along with their arguments) will
    be used by each node.

    Args:
        dispatch_id: id of current dispatch

    Returns:
        None
    """

    pass


async def run_workflow(dispatch_id: str, wait: bool = SYNC_DISPATCHES) -> RESULT_STATUS:
    """
    Plan and run the workflow by loading the result object corresponding to the
    dispatch id and retrieving essential information from it.
    Returns without changing anything if a redispatch is done of a (partially or fully)
    completed workflow with the same dispatch id.

    Args:
        dispatch_id: Dispatch id of the workflow to be run
        results_dir: Directory where the result object is stored

    Returns:
        The result object from the workflow execution
    """

    app_log.debug("Inside run_workflow.")

    # Ensure that the dispatch is run at most once
    can_run = await datasvc.ensure_dispatch(dispatch_id)

    if not can_run:
        result_info = await datasvc.dispatch.get(dispatch_id, ["status"])
        dispatch_status = result_info["status"]
        app_log.debug(f"Cannot start dispatch {dispatch_id}: current status {dispatch_status}")
        return dispatch_status

    try:
        await _plan_workflow(dispatch_id)

        if wait:
            fut = asyncio.Future()
            _futures[dispatch_id] = fut

        dispatch_status = await _submit_initial_tasks(dispatch_id)

        if wait:
            app_log.debug(f"Waiting for dispatch {dispatch_id}")
            dispatch_status = await fut
        else:
            app_log.debug(f"Running dispatch {dispatch_id} asynchronously")

    except Exception as ex:
        dispatch_status = await _handle_dispatch_exception(dispatch_id, ex)

    finally:
        if dispatch_status != RESULT_STATUS.RUNNING:
            datasvc.finalize_dispatch(dispatch_id)

    return dispatch_status


# Domain: dispatcher
async def cancel_dispatch(dispatch_id: str, task_ids: List[int] = None) -> None:
    """
    Cancel an entire dispatch or a specific set of tasks within it

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_ids: List of tasks from the lattice that are to be cancelled. Defaults to [] (entire lattice)

    Return(s)
        None

    """

    if task_ids is None:
        task_ids = []

    if not dispatch_id:
        return

    if task_ids:
        app_log.debug(f"Cancelling tasks {task_ids} in dispatch {dispatch_id}")
    else:
        task_ids = await tg_utils.get_nodes(dispatch_id)

        app_log.debug(f"Cancelling dispatch {dispatch_id}")

    await jbmgr.set_cancel_requested(dispatch_id, task_ids)
    await cancel_tasks(dispatch_id, task_ids)

    # Recursively cancel running sublattice dispatches
    attrs = await datasvc.electron.get_bulk(dispatch_id, task_ids, ["sub_dispatch_id"])
    sub_ids = list(map(lambda x: x["sub_dispatch_id"], attrs))
    for sub_dispatch_id in sub_ids:
        await cancel_dispatch(sub_dispatch_id)


def run_dispatch(dispatch_id: str) -> asyncio.Future:
    return asyncio.create_task(run_workflow(dispatch_id))


async def notify_node_status(
    dispatch_id: str, node_id: int, status: RESULT_STATUS, detail: Dict = None
):
    if detail is None:
        detail = {}

    msg = {
        "dispatch_id": dispatch_id,
        "node_id": node_id,
        "status": status,
        "detail": detail,
    }

    await _global_status_queue.put(msg)


async def _finalize_dispatch(dispatch_id: str):
    await _clear_caches(dispatch_id)
    app_log.debug(f"Removed unresolved counter for {dispatch_id}")

    incomplete_tasks = await datasvc.dispatch.get_incomplete_tasks(dispatch_id)
    failed = incomplete_tasks["failed"]
    cancelled = incomplete_tasks["cancelled"]
    if failed or cancelled:
        app_log.debug(f"Workflow {dispatch_id} cancelled or failed")
        failed_nodes = failed
        failed_nodes = map(lambda x: f"{x[0]}: {x[1]}", failed_nodes)
        failed_nodes_msg = "\n".join(failed_nodes)
        error_msg = "The following tasks failed:\n" + failed_nodes_msg
        ts = datetime.now(timezone.utc)
        status = RESULT_STATUS.FAILED if failed else RESULT_STATUS.CANCELLED
        result_update = datasvc.generate_dispatch_result(
            dispatch_id,
            status=status,
            error=error_msg,
            end_time=ts,
        )
        await datasvc.dispatch.update(dispatch_id, result_update)

    app_log.debug("8: All tasks finished running (run_planned_workflow)")

    app_log.debug("Workflow already postprocessed")

    result_info = await datasvc.dispatch.get(dispatch_id, ["status"])
    return result_info["status"]


async def _initialize_caches(dispatch_id, pending_parents, sorted_task_groups):
    for gid, indegree in pending_parents.items():
        await _pending_parents.set_pending(dispatch_id, gid, indegree)

    for gid, sorted_nodes in sorted_task_groups.items():
        await _sorted_task_groups.set_task_group(dispatch_id, gid, sorted_nodes)

    await _unresolved_tasks.set_unresolved(dispatch_id, 0)


async def _submit_initial_tasks(dispatch_id: str):
    app_log.debug("3: Inside run_planned_workflow (run_planned_workflow).")
    dispatch_result = datasvc.generate_dispatch_result(
        dispatch_id, start_time=datetime.now(timezone.utc), status=RESULT_STATUS.RUNNING
    )
    await datasvc.dispatch.update(dispatch_id, dispatch_result)

    app_log.debug(f"4: Workflow status changed to running {dispatch_id} (run_planned_workflow).")
    app_log.debug("5: Wrote lattice status to DB (run_planned_workflow).")

    initial_groups, pending_parents, sorted_task_groups = await _get_initial_tasks_and_deps(
        dispatch_id
    )

    await _initialize_caches(dispatch_id, pending_parents, sorted_task_groups)

    for gid in initial_groups:
        sorted_nodes = sorted_task_groups[gid]
        app_log.debug(f"Sorted nodes group group {gid}: {sorted_nodes}")
        await _unresolved_tasks.increment(dispatch_id, len(sorted_nodes))

    for gid in initial_groups:
        sorted_nodes = sorted_task_groups[gid]
        await _submit_task_group(dispatch_id, sorted_nodes, gid)

    return RESULT_STATUS.RUNNING


async def _handle_node_status_update(dispatch_id, node_id, node_status, detail):
    app_log.debug(f"Received node status update {node_id}: {node_status}")

    if node_status == RESULT_STATUS.RUNNING:
        return

    if node_status == RESULT_STATUS.DISPATCHING:
        sub_dispatch_id = detail["sub_dispatch_id"]
        run_dispatch(sub_dispatch_id)
        app_log.debug(f"Running sublattice dispatch {sub_dispatch_id}")

        return

    # Terminal node statuses

    if node_status == RESULT_STATUS.COMPLETED:
        next_task_groups = await _handle_completed_node(dispatch_id, node_id)
        for gid in next_task_groups:
            sorted_nodes = await _sorted_task_groups.get_task_group(dispatch_id, gid)
            await _unresolved_tasks.increment(dispatch_id, len(sorted_nodes))
            await _submit_task_group(dispatch_id, sorted_nodes, gid)

    if node_status == RESULT_STATUS.FAILED:
        await _handle_failed_node(dispatch_id, node_id)

    if node_status == RESULT_STATUS.CANCELLED:
        await _handle_cancelled_node(dispatch_id, node_id)

    # Decrement after any increments to avoid race with
    # finalize_dispatch()
    await _unresolved_tasks.decrement(dispatch_id)


async def _handle_dispatch_exception(dispatch_id: str, ex: Exception) -> RESULT_STATUS:
    error_msg = "".join(traceback.TracebackException.from_exception(ex).format())
    app_log.exception(f"Exception during _run_planned_workflow: {error_msg}")

    dispatch_result = datasvc.generate_dispatch_result(
        dispatch_id,
        end_time=datetime.now(timezone.utc),
        status=RESULT_STATUS.FAILED,
        error=error_msg,
    )

    await datasvc.dispatch.update(dispatch_id, dispatch_result)
    return RESULT_STATUS.FAILED


# msg = {
#     "dispatch_id": dispatch_id,
#     "node_id": node_id,
#     "status": status,
#     "detail": detail,
# }
async def _node_event_listener():
    app_log.debug("Starting event listener")
    while True:
        msg = await _global_status_queue.get()

        asyncio.create_task(_handle_event(msg))


async def _handle_event(msg: Dict):
    dispatch_id = msg["dispatch_id"]
    node_id = msg["node_id"]
    node_status = msg["status"]
    detail = msg["detail"]

    try:
        await _handle_node_status_update(dispatch_id, node_id, node_status, detail)

    except Exception as ex:
        dispatch_status = await _handle_dispatch_exception(dispatch_id, ex)
        await datasvc.persist_result(dispatch_id)
        fut = _futures.get(dispatch_id)
        if fut:
            fut.set_result(dispatch_status)
        return dispatch_status

    unresolved = await _unresolved_tasks.get_unresolved(dispatch_id)
    if unresolved < 1:
        app_log.debug("Finalizing dispatch")
        try:
            dispatch_status = await _finalize_dispatch(dispatch_id)
        except Exception as ex:
            dispatch_status = await _handle_dispatch_exception(dispatch_id, ex)

        finally:
            await datasvc.persist_result(dispatch_id)
            fut = _futures.get(dispatch_id)
            if fut:
                fut.set_result(dispatch_status)

        return dispatch_status


async def _clear_caches(dispatch_id: str):
    """Clean up all keys in caches."""
    await _unresolved_tasks.remove(dispatch_id)

    g_node_link = await tg_utils.get_nodes_links(dispatch_id)
    g = nx.readwrite.node_link_graph(g_node_link)

    task_groups = {g.nodes[i]["task_group_id"] for i in g.nodes}

    for gid in task_groups:
        # Clean up no longer referenced keys
        await _pending_parents.remove(dispatch_id, gid)
        await _sorted_task_groups.remove(dispatch_id, gid)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
from multiprocessing import Pipe

import socketio
import uvicorn
from fastapi import Request
from fastapi.templating import Jinja2Templates

from covalent._shared_files import logger
from covalent._shared_files.config import get_config, update_config
from covalent_dispatcher._service.app_dask import DaskCluster
from covalent_dispatcher._triggers_app import triggers_only_app  # nopycln: import
from covalent_ui.api.main import app as fastapi_app
from covalent_ui.api.main import sio
from covalent_ui.api.v1.utils.log_handler import log_config

# read env vars configuring server
COVALENT_SERVER_IFACE_ANY = os.getenv("COVALENT_SERVER_IFACE_ANY", "False").lower() in (
    "true",
    "1",
    "t",
)

WEBHOOK_PATH = "/api/webhook"
WEBAPP_PATH = "webapp/build"
STATIC_FILES = {"": WEBAPP_PATH, "/": f"{WEBAPP_PATH}/index.html"}

# Log configuration
app_log = logger.app_log
log_stack_info = logger.log_stack_info
templates = Jinja2Templates(directory=WEBAPP_PATH)


@fastapi_app.get("/{rest_of_path}")
def get_home(request: Request, rest_of_path: str):
    return templates.TemplateResponse("index.html", {"request": request})


socketio_app = socketio.ASGIApp(
    sio,
    static_files=STATIC_FILES,
)
fastapi_app.mount("/", socketio_app)

if __name__ == "__main__":
    ap = argparse.ArgumentParser()

    ap.add_argument("-p", "--port", required=False, help="Server port number.")

    ap.add_argument(
        "-d",
        "--develop",
        required=False,
        action="store_true",
        help="Start the server in developer mode.",
    )

    ap.add_argument(
        "--no-cluster",
        dest="cluster",
        action="store_false",
        required=False,
        help="Start Covalent server without Dask",
    )

    ap.add_argument(
        "--no-triggers",
        dest="no_triggers",
        action="store_true",  # this means default is False
        required=False,
        help="Start Covalent server without the Triggers server",
    )

    ap.add_argument(
        "--triggers-only",
        dest="triggers_only",
        action="store_true",  # this means default is False
        required=False,
        help="Start only the Triggers server",
    )

    ap.set_defaults(cluster=True)

    args, unknown = ap.parse_known_args()

    # port to be specified by cli
    port = int(args.port) if args.port else int(get_config("dispatcher.port"))
    host = "0.0.0.0" if COVALENT_SERVER_IFACE_ANY else get_config("dispatcher.address")

    DEBUG = args.develop is True

    # reload = True if args.develop is True else False
    RELOAD = False

    # Start dask if no-cluster flag is not specified (covalent stop auto terminates all child processes of this)
    if args.cluster:
        parent_conn, child_conn = Pipe()
        dask_cluster = DaskCluster(name="LocalDaskCluster", logger=app_log, conn=child_conn)
        dask_cluster.start()
        dask_config = parent_conn.recv()
        update_config(dask_config)

    app_name = "app:fastapi_app"
    if args.triggers_only:
        from covalent_dispatcher._triggers_app import trigger_only_router  # nopycln: import

        app_name = "app:triggers_only_app"
        fastapi_app.include_router(trigger_only_router, prefix="/api", tags=["Triggers"])
    elif args.no_triggers:
        import covalent_dispatcher._triggers_app.app as tr_app

        tr_app.disable_triggers = True

    # Start covalent main app
    uvicorn.run(
        app_name,
        host=host,
        port=port,
        debug=DEBUG,
        reload=RELOAD,
        log_config=log_config(),
    )



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from pydantic import field_validator

from ...executor.qbase import (
    BaseProcessPoolQExecutor,
    BaseQExecutor,
    BaseThreadPoolQExecutor,
    SyncBaseQExecutor,
)

SIMULATOR_DEVICES = [
    "default.qubit",
    "default.qubit.autograd",
    "default.qubit.jax",
    "default.qubit.tf",
    "default.qubit.torch",
    "default.gaussian",
    "lightning.qubit",
]


class Simulator(BaseQExecutor):
    """
    A quantum executor that uses the specified Pennylane device to execute circuits.
    Parallelizes circuit execution on the specified `device` using either threads
    or processes.

    Keyword Args:
        device: A valid string corresponding to a Pennylane device. Simulation-based
            devices (e.g. "default.qubit" and "lightning.qubit") are recommended.
            Defaults to "default.qubit" or "default.gaussian" depending on the
            decorated QNode's device.
        parallel: The type of parallelism to use. Valid values are "thread" and
            "process". Passing any other value will result in synchronous execution.
            Defaults to "thread".
        workers: The number of threads or processes to use. Defaults to 10.
        shots: The number of shots to use for the execution device. Overrides the
            :code:`shots` value from the original device if set to :code:`None` or
            a positive :code:`int`. The shots setting from the original device is
            is used by default.
    """

    device: str = "default.qubit"
    parallel: Union[bool, str] = "thread"
    workers: int = 10

    @field_validator("device")
    def validate_device(cls, device):  # pylint: disable=no-self-argument
        """
        Check that the `device` attribute is NOT a provider or hardware device.
        """
        if device not in SIMULATOR_DEVICES:
            valid_devices = ", ".join(SIMULATOR_DEVICES[::-1] + [f"or {SIMULATOR_DEVICES[-1]}"])
            raise ValueError(f"Simulator device must be {valid_devices}.")
        return device

    def batch_submit(self, qscripts_list):
        # Defer to original QNode's device type in special cases.
        if self.qelectron_info.device_name in ["default.gaussian"]:
            device = self.qelectron_info.device_name
        else:
            device = self.device

        # Select backend batching the chosen method of parallelism.
        if self.parallel == "process":
            self._backend = BaseProcessPoolQExecutor(num_processes=self.workers, device=device)
        elif self.parallel == "thread":
            self._backend = BaseThreadPoolQExecutor(num_threads=self.workers, device=device)
        else:
            self._backend = SyncBaseQExecutor(device=device)

        # Pass on server-set settings from original device.
        updates = {"device_name": device, "device_shots": self.override_shots}
        self._backend.qelectron_info = self.qelectron_info.copy(update=updates)
        self._backend.qnode_specs = self.qnode_specs.copy()

        return self._backend.batch_submit(qscripts_list)

    def batch_get_results(self, futures_list):
        return self._backend.batch_get_results(futures_list)

    _backend: BaseQExecutor = None



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Print callbacks for deploy up and deploy down commands."""

import re
from typing import Union

from rich.console import Console
from rich.status import Status


class ScrollBufferCallback:
    """Callable print callback that refreshes a buffer of length `max_lines`"""

    _complete_msg_pattern = re.compile(
        r"^(Apply complete\! Resources: \d+ added, \d+ changed, \d+ destroyed\."
        "|"
        r"Destroy complete\! Resources: \d+ destroyed\.)$"
    )

    def __init__(
        self,
        console: Console,
        console_status: Status,
        msg_template: str,
        verbose: bool,
        max_lines: int = 12,
    ):
        """Create a new scroll buffer callback.

        Args:
            console: Rich console object.
            console_status: Rich console status object.
            msg_template: Message template pre-formatted with provision or destroy message.
            verbose: Whether to print the output inline or not.
            max_lines: Number of lines in the buffer. Defaults to 5.
        """
        self.console = console
        self.console_status = console_status
        self.msg_template = msg_template
        self.verbose = verbose
        self.max_lines = max_lines
        self.buffer = []

        self._complete_msg = None

    @property
    def complete_msg(self) -> Union[str, None]:
        """Return a complete message matching:

            'Apply complete! Resources: 19 added, 0 changed, 0 destroyed.'
        or
            'Destroy complete! Resources: 19 destroyed.'

        Returns:
            The complete message, if it exists, else None.
        """
        return self._complete_msg

    def __call__(self, msg: str):
        if self.verbose:
            self._verbose_print(msg)
        else:
            self._inline_print(msg)

    def _inline_print(self, msg: str) -> None:
        """Print into a scrolling buffer of size `self.max_lines`."""
        if len(self.buffer) > self.max_lines:
            self.buffer.pop(0)
        self.buffer.append(msg)

        if self._complete_msg_pattern.match(msg):
            self._complete_msg = msg

        text = self.render_buffer()
        self.console_status.update(self.msg_template.format(text=text))

    def _verbose_print(self, msg: str) -> None:
        """Print normally, line by line."""
        print(msg.rstrip() if msg != "\n" else msg)

    def render_buffer(self, sep: str = " ") -> str:
        """Render the current buffer as a string."""
        return sep.join(self.buffer)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for logging errors, warnings, and debug statements."""

import logging
import os
import sys
import time

from .config import get_config

logging.Formatter.converter = time.gmtime

app_log = logging.getLogger(__name__)

log_level = get_config("sdk.log_level").upper()
app_log.setLevel(log_level)

# Set the format
stream_formatter = logging.Formatter(
    "[%(asctime)s] " + "[%(levelname)s] " + "%(filename)s: " + "Line "
    "%(lineno)s in %(funcName)s: " + "%(message)s"
)

stream_handler = logging.StreamHandler()
stream_handler.setLevel(log_level)
stream_handler.setFormatter(stream_formatter)
stream_handler.stream = sys.stdout

app_log.addHandler(stream_handler)
app_log.propagate = False

# Show stack traces
log_stack_info = os.environ.get("LOGSTACK", "TRUE").upper() == "TRUE"



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for transport graph operations."""

from collections import deque
from typing import Callable, List

import networkx as nx

from covalent._shared_files import logger
from covalent._shared_files.util_classes import RESULT_STATUS

from .asset import copy_asset, copy_asset_meta
from .electron import ASSET_KEYS, METADATA_KEYS
from .tg import _TransportGraph

app_log = logger.app_log


class TransportGraphOps:
    def __init__(self, tg: _TransportGraph):
        self.tg = tg
        self._status_map = {1: True, -1: False}

        self._default_node_attrs = {
            "start_time": None,
            "end_time": None,
            "status": RESULT_STATUS.NEW_OBJECT,
            # This will be overwritten by the SDK
            # "output": None,
            # "error": "",
            # "stdout": "",
            # "stderr": "",
        }

    @staticmethod
    def _flag_successors(A: nx.MultiDiGraph, node_statuses: dict, starting_node: int):
        """Flag all successors of a node (including the node itself)."""
        nodes_to_invalidate = [starting_node]
        for node, successors in nx.bfs_successors(A, starting_node):
            nodes_to_invalidate.extend(iter(successors))
        for node in nodes_to_invalidate:
            node_statuses[node] = -1

    @staticmethod
    def is_same_node(A: nx.MultiDiGraph, B: nx.MultiDiGraph, node: int) -> bool:
        """Check if the node attributes are the same in both graphs."""
        return A.nodes[node] == B.nodes[node]

    @staticmethod
    def is_same_edge_attributes(
        A: nx.MultiDiGraph, B: nx.MultiDiGraph, parent: int, node: int
    ) -> bool:
        """Check if the edge attributes are the same in both graphs."""
        return A.adj[parent][node] == B.adj[parent][node]

    def copy_nodes_from(
        self,
        tg: _TransportGraph,
        nodes,
        *,
        copy_metadata: bool = True,
        defer_copy_objects: bool = False,
    ) -> List:
        """Copy nodes from the transport graph in the argument."""

        assets_to_copy = []

        for n in nodes:
            old_node = tg.get_node(n)
            old_status = tg.get_node_value(n, "status")

            if copy_metadata and old_status == RESULT_STATUS.COMPLETED:
                # Only previously completed nodes can actually be
                # reused

                for k in METADATA_KEYS:
                    app_log.debug(f"Copying metadata {k} for node {n}")
                    v = tg.get_node_value(n, k)
                    if k == "status":
                        # This will cause the dispatcher to skip
                        # re-running the node.
                        v = RESULT_STATUS.PENDING_REUSE
                    self.tg.set_node_value(n, k, v)

            # TODO: Use the ElectronAssets link table as the source of
            # truth instead of these hardcoded values
            for k in ASSET_KEYS:
                # Copy asset metadata
                app_log.debug(f"Copying asset {k} for node {n}")
                with old_node.session() as session:
                    old = old_node.get_asset(k, session)
                    new = self.tg.get_node(n).get_asset(k, session)
                    copy_asset_meta(session, old, new)
                    assets_to_copy.append((old, new))

        # Now perform all data copy operations (this could be slow)
        if not defer_copy_objects:
            for item in assets_to_copy:
                src, dest = item
                copy_asset(src, dest)

        # Return the assets to copy at a later time
        return assets_to_copy

    @staticmethod
    def _cmp_name_and_pval(A: nx.MultiDiGraph, B: nx.MultiDiGraph, node: int) -> bool:
        """Default node comparison function for diffing transport graphs.

        Two nodes are considered the "same" if either:
        * both are parameter nodes and have the same hash
        * both are function nodes, have the same name,
          and neither is marked PENDING_REPLACEMENT.
        """

        name_A = A.nodes[node]["name"]
        name_B = B.nodes[node]["name"]

        if name_A != name_B:
            return False

        status_A = A.nodes[node]["status"]
        status_B = B.nodes[node]["status"]
        if (
            status_A == RESULT_STATUS.PENDING_REPLACEMENT
            or status_B == RESULT_STATUS.PENDING_REPLACEMENT
        ):
            return False

        # Same name -- remaining case to check is if both are
        # parameters.  Compare parameter value hashes.
        val_hash_A = A.nodes[node].get("value", None)
        val_hash_B = B.nodes[node].get("value", None)

        return val_hash_A == val_hash_B

    def _max_cbms(
        self,
        A: nx.MultiDiGraph,
        B: nx.MultiDiGraph,
        node_cmp: Callable = None,
        edge_cmp: Callable = None,
    ):
        """Computes a "maximum backward-maximal common subgraph" (cbms)
        Args:
            A: nx.MultiDiGraph
            B: nx.MultiDiGraph
            node_cmp: An optional function for comparing node attributes in A and B.
                    Defaults to testing for equality of the attribute dictionaries
            edge_cmp: An optional function for comparing the edges between two nodes.
                    Defaults to checking that the two sets of edges have the same attributes
        Returns: A_node_status, B_node_status, where each is a dictionary
            `{node: True/False}` where True means reusable.
        Performs a modified BFS of A and B.
        """
        if node_cmp is None:
            node_cmp = self.is_same_node
        if edge_cmp is None:
            edge_cmp = self.is_same_edge_attributes

        A_node_status = {node_id: 0 for node_id in A.nodes}
        B_node_status = {node_id: 0 for node_id in B.nodes}
        app_log.debug(f"A node status: {A_node_status}")
        app_log.debug(f"B node status: {B_node_status}")

        virtual_root = -1

        if virtual_root in A.nodes or virtual_root in B.nodes:
            raise RuntimeError(f"Encountered forbidden node: {virtual_root}")

        assert virtual_root not in B.nodes

        nodes_to_visit = deque()
        nodes_to_visit.appendleft(virtual_root)

        # Add a temporary root
        A_parentless_nodes = [node for node, deg in A.in_degree() if deg == 0]
        B_parentless_nodes = [node for node, deg in B.in_degree() if deg == 0]
        for node_id in A_parentless_nodes:
            A.add_edge(virtual_root, node_id)

        for node_id in B_parentless_nodes:
            B.add_edge(virtual_root, node_id)

        # Assume inductively that predecessors subgraphs are the same;
        # this is satisfied for the root
        while nodes_to_visit:
            current_node = nodes_to_visit.pop()

            app_log.debug(f"Visiting node {current_node}")
            for y in A.adj[current_node]:
                # Don't process already failed nodes
                if A_node_status[y] == -1:
                    continue

                # Check if y is a valid child of current_node in B
                if y not in B.adj[current_node]:
                    app_log.debug(f"A: {y} not adjacent to node {current_node} in B")
                    self._flag_successors(A, A_node_status, y)
                    continue

                if B_node_status[y] == -1:
                    app_log.debug(f"A: Node {y} is marked as failed in B")
                    self._flag_successors(A, A_node_status, y)
                    continue

                # Compare edges
                if not edge_cmp(A, B, current_node, y):
                    app_log.debug(f"Edges between {current_node} and {y} differ")
                    self._flag_successors(A, A_node_status, y)
                    self._flag_successors(B, B_node_status, y)
                    continue

                # Compare nodes
                if not node_cmp(A, B, y):
                    app_log.debug(f"Attributes of node {y} differ:")
                    app_log.debug(f"A[y] = {A.nodes[y]}")
                    app_log.debug(f"B[y] = {B.nodes[y]}")
                    self._flag_successors(A, A_node_status, y)
                    self._flag_successors(B, B_node_status, y)
                    continue

                # Predecessors subgraphs of y are the same in A and B, so
                # enqueue y if it hasn't already been visited
                assert A_node_status[y] != -1
                if A_node_status[y] == 0:
                    A_node_status[y] = 1
                    B_node_status[y] = 1
                    app_log.debug(f"Enqueueing node {y}")
                    nodes_to_visit.appendleft(y)

            # Prune children of current_node in B that aren't valid children in A
            for y in B.adj[current_node]:
                if B_node_status[y] == -1:
                    continue
                if y not in A.adj[current_node]:
                    app_log.debug(f"B: {y} not adjacent to node {current_node} in A")
                    self._flag_successors(B, B_node_status, y)
                    continue
                if B_node_status[y] == -1:
                    app_log.debug(f"B: Node {y} is marked as failed in A")
                    self._flag_successors(B, B_node_status, y)

        A.remove_node(-1)
        B.remove_node(-1)

        app_log.debug(f"A node status: {A_node_status}")
        app_log.debug(f"B node status: {B_node_status}")

        for k, v in A_node_status.items():
            A_node_status[k] = self._status_map[v]
        for k, v in B_node_status.items():
            B_node_status[k] = self._status_map[v]
        return A_node_status, B_node_status

    def get_reusable_nodes(self, tg_new: _TransportGraph) -> List[int]:
        """Find which nodes are common between the current graph and a new graph."""
        A = self.tg.get_internal_graph_copy()
        B = tg_new.get_internal_graph_copy()

        # inject parameter value checksums
        for node_id in A.nodes:
            node = self.tg.get_node(node_id)
            with node.session() as session:
                value_asset = node.get_asset("value", session)
            value_hash = value_asset.digest
            A.nodes[node_id]["value"] = value_hash

        for node_id in B.nodes:
            node = tg_new.get_node(node_id)
            with node.session() as session:
                value_asset = node.get_asset("value", session)
            value_hash = value_asset.digest
            B.nodes[node_id]["value"] = value_hash

        status_A, _ = self._max_cbms(A, B, node_cmp=self._cmp_name_and_pval)
        return [k for k, v in status_A.items() if v]

    def reset_nodes(self):
        """Reset all nodes to be replaced."""
        for node_id in self.tg._graph.nodes:
            status = self.tg.get_node_value(node_id, "status")
            if status == RESULT_STATUS.PENDING_REPLACEMENT:
                self._reset_node(node_id)

    def _reset_node(self, node_id: int) -> None:
        """Reset node values to starting state."""
        node_name = self.tg.get_node_value(node_id, "name")

        for node_attr, default_val in self._default_node_attrs.items():
            self.tg.set_node_value(node_id, node_attr, default_val)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from threading import Thread
from typing import Any, Dict, List, Optional, Sequence, Union

import orjson
import pennylane as qml
from mpire import WorkerPool
from pydantic import (  # pylint: disable=no-name-in-module
    BaseModel,
    ConfigDict,
    Field,
    model_validator,
)

from .._shared_files.qinfo import QElectronInfo, QNodeSpecs

__all__ = [
    "BaseQExecutor",
    "BaseProcessPoolQExecutor",
    "AsyncBaseQExecutor",
    "BaseThreadPoolQExecutor",
]

SHOTS_DEFAULT = -1


def orjson_dumps(v, *, default):
    return orjson.dumps(v, default=default).decode()  # pylint: disable=no-member


@lru_cache
def get_process_pool(num_processes=None):
    return WorkerPool(n_jobs=num_processes)


@lru_cache
def get_thread_pool(max_workers=None):
    return ThreadPoolExecutor(max_workers=max_workers)


@lru_cache
def get_asyncio_event_loop():
    """
    Returns an asyncio event loop running in a separate thread.
    """

    def _run_loop(_loop):
        asyncio.set_event_loop(_loop)
        _loop.run_forever()

    loop = asyncio.new_event_loop()
    thread = Thread(target=_run_loop, args=(loop,), daemon=True)
    thread.start()

    # Create function attribute so reference to thread is not lost.
    get_asyncio_event_loop.thread = thread

    return loop


class BaseQExecutor(ABC, BaseModel):
    """
    Base class for all Quantum Executors.
    """

    shots: Union[None, int, Sequence[int], Sequence[Union[int, Sequence[int]]]] = SHOTS_DEFAULT
    shots_converter: Optional[type] = None
    persist_data: bool = True

    # Executors need to contain certain information about original QNode, in order
    # to produce correct results. These attributes below contain that information.
    # They are set inside the `QServer` and will be `None` client-side.
    qelectron_info: Optional[QElectronInfo] = None
    qnode_specs: Optional[QNodeSpecs] = None

    @property
    def override_shots(self) -> Union[int, None]:
        """
        Fallback to the QNode device's shots if no user-specified shots on executor.
        """

        if self.shots is SHOTS_DEFAULT:
            # No user-specified shots. Use the original QNode device's shots instead.
            shots = self.qelectron_info.device_shots
            shots_converter = self.qelectron_info.device_shots_type
            return shots_converter(shots) if shots_converter is not None else shots
        if self.shots is None:
            # User has specified `shots=None` on executor.
            return None

        if isinstance(self.shots, Sequence) and self.shots_converter is not None:
            return self.shots_converter(self.shots)

        # User has specified `shots` as an int.
        return self.shots

    model_config = ConfigDict(extra="allow")

    @model_validator(mode="before")
    @classmethod
    def set_name(cls, values):
        # pylint: disable=no-self-argument
        # Set the `name` attribute to the class name
        values["name"] = cls.__name__
        return values

    @abstractmethod
    def batch_submit(self, qscripts_list):
        raise NotImplementedError

    @abstractmethod
    def batch_get_results(self, futures_list):
        raise NotImplementedError

    def run_circuit(self, qscript, device, result_obj: "QCResult") -> "QCResult":
        start_time = time.perf_counter()
        results = qml.execute([qscript], device, gradient_fn="best")
        end_time = time.perf_counter()

        result_obj.results = results
        result_obj.execution_time = end_time - start_time

        return result_obj

    def dict(self, *args, **kwargs):
        dict_ = super().model_dump(*args, **kwargs)

        # Ensure shots is a hashable value.
        shots = dict_.get("shots")
        if isinstance(shots, Sequence):
            dict_["shots"] = tuple(shots)

            # Set shots converter to recover original sequence type.
            shots_converter = dict_.get("shots_converter")
            if shots_converter is None:
                dict_["shots_converter"] = type(shots)

        return dict_


class QCResult(BaseModel):
    """
    Container for results from `run_circuit` methods. Standardizes output and allows
    metadata to be updated at various points.
    """

    results: Optional[Any] = None
    execution_time: float = None
    metadata: Dict[str, Any] = Field(default_factory=lambda: {"execution_metadata": []})

    def expand(self) -> List["QCResult"]:
        """
        Expand result object into a list of result objects, one for each execution.
        """
        result_objs = []
        for i, result in enumerate(self.results):
            # Copy other non-execution metadata.
            _result_obj = QCResult(
                results=[result], execution_time=self.execution_time, metadata={}
            )

            # Handle single and multi-component metadata.
            execution_metadata = self.metadata["execution_metadata"]
            if len(self.metadata["execution_metadata"]) > 0:
                execution_metadata = execution_metadata[i]

            # Populate corresponding metadata.
            _result_obj.metadata.update(
                execution_metadata=[execution_metadata],
                device_name=self.metadata["device_name"],
                executor_name=self.metadata["executor_name"],
                executor_backend_name=self.metadata["executor_backend_name"],
            )

            result_objs.append(_result_obj)

        return result_objs

    @classmethod
    def with_metadata(cls, *, device_name: str, executor: BaseQExecutor):
        """
        Create a blank instance with pre-set metadata.
        """
        result_obj = cls()
        backend_name = executor.backend if hasattr(executor, "backend") else ""
        result_obj.metadata.update(
            device_name=device_name,
            executor_name=executor.__class__.__name__,
            executor_backend_name=backend_name,
        )
        return result_obj


class SyncBaseQExecutor(BaseQExecutor):
    device: Optional[str] = "default.qubit"

    def run_all_circuits(self, qscripts_list) -> List[QCResult]:
        result_objs: List[QCResult] = []

        for qscript in qscripts_list:
            dev = qml.device(
                self.device,
                wires=self.qelectron_info.device_wires,
                shots=self.qelectron_info.device_shots,
            )

            result_obj = QCResult.with_metadata(device_name=dev.short_name, executor=self)
            result_obj = self.run_circuit(qscript, dev, result_obj)
            result_objs.append(result_obj)

        return result_objs

    def batch_submit(self, qscripts_list):
        # Offload execution of all circuits to the same thread
        # so that the qserver isn't blocked by their completion.
        pool = get_thread_pool()
        fut = pool.submit(self.run_all_circuits, qscripts_list)
        dummy_futures = [fut] * len(qscripts_list)
        return dummy_futures

    def batch_get_results(self, futures_list):
        return futures_list[0].result()


class AsyncBaseQExecutor(BaseQExecutor):
    """
    Executor that uses `asyncio` to handle multiple job submissions
    """

    device: Optional[str] = "default.qubit"

    def batch_submit(self, qscripts_list):
        futures = []
        loop = get_asyncio_event_loop()
        for qscript in qscripts_list:
            dev = qml.device(
                self.device,
                wires=self.qelectron_info.device_wires,
                shots=self.qelectron_info.device_shots,
            )

            result_obj = QCResult.with_metadata(
                device_name=dev.short_name,
                executor=self,
            )
            fut = loop.create_task(self.run_circuit(qscript, dev, result_obj))
            futures.append(fut)

        return futures

    def batch_get_results(self, futures_list: List):
        loop = get_asyncio_event_loop()
        task = asyncio.run_coroutine_threadsafe(self._get_result(futures_list), loop)
        return task.result()

    async def _get_result(self, futures_list: List) -> List[QCResult]:
        return await asyncio.gather(*futures_list)

    async def run_circuit(self, qscript, device, result_obj) -> QCResult:
        await asyncio.sleep(0)
        start_time = time.perf_counter()
        results = qml.execute([qscript], device, gradient_fn="best")
        end_time = time.perf_counter()

        result_obj.results = results
        result_obj.execution_time = end_time - start_time

        return result_obj


class BaseProcessPoolQExecutor(BaseQExecutor):
    device: Optional[str] = "default.qubit"
    num_processes: int = 10

    def batch_submit(self, qscripts_list):
        pool = get_process_pool(self.num_processes)

        futures = []
        for qscript in qscripts_list:
            dev = qml.device(
                self.device,
                wires=self.qelectron_info.device_wires,
                shots=self.qelectron_info.device_shots,
            )

            result_obj = QCResult.with_metadata(
                device_name=dev.short_name,
                executor=self,
            )
            fut = pool.apply_async(self.run_circuit, args=(qscript, dev, result_obj))
            futures.append(fut)

        return futures

    def batch_get_results(self, futures_list: List) -> List[QCResult]:
        return [fut.get() for fut in futures_list]


class BaseThreadPoolQExecutor(BaseQExecutor):
    device: Optional[str] = "default.qubit"
    num_threads: int = 10

    def batch_submit(self, qscripts_list):
        pool = get_thread_pool(self.num_threads)

        futures = []
        for qscript in qscripts_list:
            dev = qml.device(
                self.device,
                wires=self.qelectron_info.device_wires,
                shots=self.qelectron_info.device_shots,
            )

            result_obj = QCResult.with_metadata(
                device_name=dev.short_name,
                executor=self,
            )
            fut = pool.submit(self.run_circuit, qscript, dev, result_obj)
            futures.append(fut)

        return futures

    def batch_get_results(self, futures_list: List) -> List[QCResult]:
        return [fut.result() for fut in futures_list]



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Import workflow utilities."""

from .depsbash import DepsBash
from .depscall import DepsCall
from .depsmodule import DepsModule
from .depspip import DepsPip
from .electron import electron
from .lattice import lattice
from .lepton import Lepton
from .transport import TransportableObject



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import json
import os
from pathlib import Path
from typing import Any, Optional, Tuple

import cloudpickle

from covalent._shared_files.config import get_config
from covalent._shared_files.schemas import electron, lattice, result
from covalent._workflow.transport import TransportableObject

from .base import BaseProvider, Digest

BLOCK_SIZE = 65536
ALGORITHM = "sha1"

WORKFLOW_ASSET_FILENAME_MAP = result.ASSET_FILENAME_MAP.copy()
WORKFLOW_ASSET_FILENAME_MAP.update(lattice.ASSET_FILENAME_MAP)
ELECTRON_ASSET_FILENAME_MAP = electron.ASSET_FILENAME_MAP.copy()


# Moved from write_result_to_db.py


class InvalidFileExtension(Exception):
    """
    Exception to raise when an invalid file extension is encountered
    """

    pass


class LocalProvider(BaseProvider):
    scheme = "file"

    def __init__(self):
        self.base_path = get_config("dispatcher.results_dir")

    def digest(self, bucket_name: str, object_key: str) -> Digest:
        path = os.path.join(bucket_name, object_key)
        h = hashlib.new(ALGORITHM)
        with open(path, "rb") as f:
            buf = f.read(BLOCK_SIZE)
            while len(buf) > 0:
                h.update(buf)
                buf = f.read(BLOCK_SIZE)

        return Digest(algorithm=ALGORITHM, hexdigest=h.hexdigest())

    def size(self, bucket_name: str, object_key: str) -> int:
        path = os.path.join(bucket_name, object_key)

        try:
            return os.path.getsize(path)
        except OSError:
            return 0

    def get_uri_components(
        self, dispatch_id: str, node_id: Optional[int], asset_key: str
    ) -> Tuple[str, str]:
        """Compute storage_path and object_key for a workflow asset.

        Args:
            dispatch_id: The workflow dispatch id
            node_id: The electron's node id or `None` if the asset has workflow scope.
            asset_key: The key describing the asset.

        Returns:
            storage_path, object_key

        The semantics `storage_path` and `object_key` may differ
        slightly between backends but are constrained by the requirement that
        `{scheme}://{storage_path}/{object_key}` is a valid URI for
        the asset.

        """
        storage_path = os.path.join(self.base_path, dispatch_id)

        if node_id is not None:
            storage_path = os.path.join(storage_path, f"node_{node_id}")
            object_key = ELECTRON_ASSET_FILENAME_MAP[asset_key]
        else:
            object_key = WORKFLOW_ASSET_FILENAME_MAP[asset_key]

        os.makedirs(storage_path, exist_ok=True)

        return storage_path, object_key

    def store_file(self, storage_path: str, filename: str, data: Any = None) -> Tuple[Digest, int]:
        """This function writes data corresponding to the filepaths in the DB."""

        if data is None:
            return Digest(algorithm="sha1", hexdigest=""), 0

        if filename.endswith(".pkl"):
            with open(Path(storage_path) / filename, "wb") as f:
                cloudpickle.dump(data, f)

        elif filename.endswith(".log") or filename.endswith(".txt"):
            if data is None:
                data = ""

            if not isinstance(data, str):
                raise InvalidFileExtension("Data must be string type.")

            with open(Path(storage_path) / filename, "w+") as f:
                f.write(data)

        elif filename.endswith(".tobj"):
            with open(Path(storage_path) / filename, "wb") as f:
                f.write(data.serialize())

        elif filename.endswith(".json"):
            with open(Path(storage_path) / filename, "w") as f:
                json.dump(data, f)

        elif filename.endswith(".mdb"):
            with open(Path(storage_path) / filename, "wb") as f:
                f.write(data)
        else:
            raise InvalidFileExtension("The file extension is not supported.")

        digest = self.digest(bucket_name=storage_path, object_key=filename)
        size = self.size(bucket_name=storage_path, object_key=filename)
        return digest, size

    def load_file(self, storage_path: str, filename: str) -> Any:
        """This function loads data for the filenames in the DB."""

        if filename.endswith(".pkl"):
            with open(Path(storage_path) / filename, "rb") as f:
                data = cloudpickle.load(f)

        elif filename.endswith(".log") or filename.endswith(".txt"):
            with open(Path(storage_path) / filename, "r") as f:
                data = f.read()

        elif filename.endswith(".tobj"):
            with open(Path(storage_path) / filename, "rb") as f:
                data = TransportableObject.deserialize(f.read())

        elif filename.endswith(".json"):
            with open(Path(storage_path) / filename, "r") as f:
                data = json.load(f)

        elif filename.endswith(".mdb"):
            with open(Path(storage_path) / filename, "rb") as f:
                data = f.read()

        return data


local_store = LocalProvider()



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from types import ModuleType
from typing import Union

from .depscall import DepsCall
from .transportable_object import TransportableObject


class DepsModule(DepsCall):
    """
    Python modules to be imported in an electron's execution environment.

    This is only used as a vehicle to send the module by reference to the
    to the right place of serialization where it will instead be pickled by value.

    Deps class to encapsulate python modules to be
    imported in the same execution environment as the electron.

    Note: This subclasses the DepsCall class due to its pre-existing
    infrastructure integrations, and not because of its logical functionality.

    Attributes:
        module_name: A string containing the name of the module to be imported.
    """

    def __init__(self, module: Union[str, ModuleType]):
        if isinstance(module, str):
            # Import the module on the client side
            module = importlib.import_module(module)

        # Temporarily pickling the module by reference
        # so that it can be pickled by value when serializing
        # the transport graph.
        self.pickled_module = TransportableObject(module)

        super().__init__()

    def short_name(self):
        """Returns the short name of this class."""
        return "depsmodule"



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions to transform ResultSchema -> Result"""

import json
import os

from sqlalchemy.orm import Session

from covalent._shared_files.config import get_config
from covalent._shared_files.schemas.lattice import (
    LATTICE_COVA_IMPORTS_FILENAME,
    LATTICE_DOCSTRING_FILENAME,
    LATTICE_ERROR_FILENAME,
    LATTICE_FUNCTION_FILENAME,
    LATTICE_FUNCTION_STRING_FILENAME,
    LATTICE_HOOKS_FILENAME,
    LATTICE_INPUTS_FILENAME,
    LATTICE_LATTICE_IMPORTS_FILENAME,
    LATTICE_NAMED_ARGS_FILENAME,
    LATTICE_NAMED_KWARGS_FILENAME,
    LATTICE_RESULTS_FILENAME,
    LATTICE_STORAGE_TYPE,
    LatticeAssets,
    LatticeSchema,
)
from covalent_dispatcher._dal.asset import Asset
from covalent_dispatcher._dal.lattice import Lattice
from covalent_dispatcher._object_store.local import BaseProvider


def _get_lattice_meta(lat: LatticeSchema, storage_path) -> dict:
    results_dir = os.environ.get("COVALENT_DATA_DIR") or get_config("dispatcher.results_dir")
    kwargs = {
        "results_dir": results_dir,  # Needed for current executors
        "storage_path": storage_path,
        "storage_type": LATTICE_STORAGE_TYPE,
        "name": lat.metadata.name,
        "python_version": lat.metadata.python_version,
        "covalent_version": lat.metadata.covalent_version,
        "executor": lat.metadata.executor,
        "executor_data": json.dumps(lat.metadata.executor_data),
        "workflow_executor": lat.metadata.workflow_executor,
        "workflow_executor_data": json.dumps(lat.metadata.workflow_executor_data),
    }
    num_nodes = len(lat.transport_graph.nodes)
    db_kwargs = {
        "electron_num": num_nodes,
        "completed_electron_num": 0,
    }
    kwargs.update(db_kwargs)

    legacy_kwargs = {
        "docstring_filename": LATTICE_DOCSTRING_FILENAME,
        "function_filename": LATTICE_FUNCTION_FILENAME,
        "function_string_filename": LATTICE_FUNCTION_STRING_FILENAME,
        "error_filename": LATTICE_ERROR_FILENAME,
        "inputs_filename": LATTICE_INPUTS_FILENAME,
        "named_args_filename": LATTICE_NAMED_ARGS_FILENAME,
        "named_kwargs_filename": LATTICE_NAMED_KWARGS_FILENAME,
        "results_filename": LATTICE_RESULTS_FILENAME,
        "hooks_filename": LATTICE_HOOKS_FILENAME,
        "cova_imports_filename": LATTICE_COVA_IMPORTS_FILENAME,
        "lattice_imports_filename": LATTICE_LATTICE_IMPORTS_FILENAME,
    }
    kwargs.update(legacy_kwargs)
    return kwargs


def import_lattice_assets(
    session: Session,
    dispatch_id: str,
    lat: LatticeSchema,
    record: Lattice,
    object_store: BaseProvider,
) -> LatticeAssets:
    """Insert asset records and populate the asset link table"""
    asset_ids = {}

    # Register built-in assets
    for asset_key, asset in lat.assets:
        storage_path, object_key = object_store.get_uri_components(
            dispatch_id=dispatch_id,
            node_id=None,
            asset_key=asset_key,
        )

        local_uri = os.path.join(storage_path, object_key)

        asset_kwargs = {
            "storage_type": object_store.scheme,
            "storage_path": storage_path,
            "object_key": object_key,
            "digest_alg": asset.digest_alg,
            "digest": asset.digest,
            "remote_uri": asset.uri,
            "size": asset.size,
        }
        asset_ids[asset_key] = Asset.create(session, insert_kwargs=asset_kwargs, flush=False)

        # Send this back to the client
        asset.digest = None
        asset.remote_uri = f"file://{local_uri}"

    # Register custom assets
    if lat.custom_assets:
        for asset_key, asset in lat.custom_assets.items():
            object_key = f"{asset_key}.data"
            local_uri = os.path.join(storage_path, object_key)

            asset_kwargs = {
                "storage_type": object_store.scheme,
                "storage_path": storage_path,
                "object_key": object_key,
                "digest_alg": asset.digest_alg,
                "digest": asset.digest,
                "remote_uri": asset.uri,
                "size": asset.size,
            }
            asset_ids[asset_key] = Asset.create(session, insert_kwargs=asset_kwargs, flush=False)

            # Send this back to the client
            asset.remote_uri = f"file://{local_uri}" if asset.digest else ""
            asset.digest = None

    session.flush()

    # Write asset records to DB
    session.flush()

    lattice_asset_links = [
        record.associate_asset(session, key, asset_rec.id) for key, asset_rec in asset_ids.items()
    ]
    session.flush()

    return lat.assets



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""v12

Revision ID: 26d71848a404
Revises: 9b9d58f02985
Create Date: 2022-09-08 10:46:11.140472

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
# pragma: allowlist nextline secret
revision = "26d71848a404"
# pragma: allowlist nextline secret
down_revision = "9b9d58f02985"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("lattices", schema=None) as batch_op:
        batch_op.add_column(sa.Column("docstring_filename", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("deps_filename", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("call_before_filename", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("call_after_filename", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("cova_imports_filename", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("lattice_imports_filename", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("results_dir", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("root_dispatch_id", sa.String(length=64), nullable=True))

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("lattices", schema=None) as batch_op:
        batch_op.drop_column("root_dispatch_id")
        batch_op.drop_column("results_dir")
        batch_op.drop_column("lattice_imports_filename")
        batch_op.drop_column("cova_imports_filename")
        batch_op.drop_column("call_after_filename")
        batch_op.drop_column("call_before_filename")
        batch_op.drop_column("deps_filename")
        batch_op.drop_column("docstring_filename")

    # ### end Alembic commands ###



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions to transform Lattice -> LatticeSchema"""


from covalent._shared_files.schemas.asset import AssetSchema
from covalent._shared_files.schemas.lattice import LatticeAssets, LatticeMetadata, LatticeSchema
from covalent_dispatcher._dal.lattice import ASSET_KEYS, METADATA_KEYS, Lattice

from .tg import export_transport_graph


def _export_lattice_meta(lat: Lattice) -> LatticeMetadata:
    metadata_kwargs = {key: lat.get_value(key, None, refresh=False) for key in METADATA_KEYS}
    return LatticeMetadata(**metadata_kwargs)


def _export_lattice_assets(lat: Lattice) -> LatticeAssets:
    manifests = {}
    for asset_key in ASSET_KEYS:
        asset = lat.assets[asset_key]
        size = asset.size
        digest_alg = asset.digest_alg
        digest = asset.digest
        scheme = asset.storage_type.value
        remote_uri = f"{scheme}://{asset.storage_path}/{asset.object_key}"
        manifests[asset_key] = AssetSchema(
            remote_uri=remote_uri, size=size, digest_alg=digest_alg, digest=digest
        )
    return LatticeAssets(**manifests)


def export_lattice(lat: Lattice) -> LatticeSchema:
    metadata = _export_lattice_meta(lat)
    assets = _export_lattice_assets(lat)
    transport_graph = export_transport_graph(lat.transport_graph)
    return LatticeSchema(metadata=metadata, assets=assets, transport_graph=transport_graph)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilties to transfer data between Covalent and compute backends
"""

import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Dict

from covalent._shared_files import logger
from covalent._shared_files.schemas.asset import AssetUpdate

from ..._dal.result import get_result_object as get_result_object

app_log = logger.app_log
am_pool = ThreadPoolExecutor()


# Consumed by Runner
async def upload_asset_for_nodes(dispatch_id: str, key: str, dest_uris: dict):
    """Typical keys: "output", "deps", "call_before", "call_after", "function"""

    loop = asyncio.get_running_loop()
    await loop.run_in_executor(
        am_pool,
        upload_asset_for_nodes_sync,
        dispatch_id,
        key,
        dest_uris,
    )


def upload_asset_for_nodes_sync(dispatch_id: str, key: str, dest_uris: dict):
    result_object = get_result_object(dispatch_id, bare=True)
    tg = result_object.lattice.transport_graph

    uploads_by_node = {}

    with result_object.session() as session:
        for node_id, dest_uri in dest_uris.items():
            if dest_uri:
                node = tg.get_node(node_id, session)
                asset = node.get_asset(key, session)
                uploads_by_node[node_id] = (asset, dest_uri)

    for _, pending_upload in uploads_by_node.items():
        asset, dest_uri = pending_upload
        asset.upload(dest_uri)


async def download_assets_for_node(
    dispatch_id: str, node_id: int, asset_updates: Dict[str, AssetUpdate]
):
    # Keys for src_uris: "output", "stdout", "stderr"

    loop = asyncio.get_running_loop()
    await loop.run_in_executor(
        am_pool,
        download_assets_for_node_sync,
        dispatch_id,
        node_id,
        asset_updates,
    )


def download_assets_for_node_sync(
    dispatch_id: str, node_id: int, asset_updates: Dict[str, AssetUpdate]
):
    result_object = get_result_object(dispatch_id, bare=True)
    tg = result_object.lattice.transport_graph
    node = tg.get_node(node_id)

    db_updates = {}

    # Mapping from asset key to (Asset, remote uri)
    assets_to_download = {}

    # Prepare asset metadata update; prune empty fields
    with result_object.session() as session:
        for key in asset_updates:
            asset_update = asset_updates[key]
            if asset_update.remote_uri:
                asset = node.get_asset(key, session)
                assets_to_download[key] = (asset, asset_update.remote_uri)
            # Prune unset fields
            db_updates[key] = {attr: val for attr, val in asset_update if val is not None}

        node.update_assets(db_updates, session)

    for key, pending_download in assets_to_download.items():
        asset, remote_uri = pending_download
        asset.download(remote_uri)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions to transform ResultSchema -> Result"""

import os
from datetime import datetime
from typing import List, Optional, Tuple

from sqlalchemy.orm import Session

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.schemas.lattice import LatticeSchema
from covalent._shared_files.schemas.result import ResultAssets, ResultSchema
from covalent._shared_files.utils import format_server_url
from covalent_dispatcher._dal.asset import Asset
from covalent_dispatcher._dal.electron import ElectronMeta
from covalent_dispatcher._dal.job import Job
from covalent_dispatcher._dal.result import Result, ResultMeta
from covalent_dispatcher._object_store.local import BaseProvider, local_store

from ..asset import copy_asset_meta
from ..tg_ops import TransportGraphOps
from ..utils.uri_filters import AssetScope, URIFilterPolicy, filter_asset_uri
from .lattice import _get_lattice_meta, import_lattice_assets
from .tg import import_transport_graph

SERVER_URL = format_server_url(get_config("dispatcher.address"), get_config("dispatcher.port"))

URI_FILTER_POLICY = URIFilterPolicy[get_config("dispatcher.data_uri_filter_policy")]

app_log = logger.app_log


def import_result(
    res: ResultSchema,
    base_path: str,
    electron_id: Optional[int],
) -> ResultSchema:
    """Imports a ResultSchema into the DB"""

    dispatch_id = res.metadata.dispatch_id

    # If result already exists in the DB, it was previously registered
    # as a sublattice dispatch; in that case, just connect it to its
    # parent electron.
    with Result.session() as session:
        records = ResultMeta.get(
            session,
            fields={"id", "dispatch_id"},
            equality_filters={"dispatch_id": dispatch_id},
            membership_filters={},
        )
        if len(records) > 0:
            return _connect_result_to_electron(session, res, electron_id)

    # Main case: insert new lattice, electron, edge, and job records

    storage_path = os.path.join(base_path, dispatch_id)

    lattice_record_kwargs = _get_result_meta(res, storage_path, electron_id)
    lattice_record_kwargs.update(_get_lattice_meta(res.lattice, storage_path))

    with Result.session() as session:
        st = datetime.now()
        lattice_row = ResultMeta.create(session, insert_kwargs=lattice_record_kwargs, flush=True)
        res_record = Result(session, lattice_row, True)
        res_assets = import_result_assets(session, res, res_record, local_store)

        lat_assets = import_lattice_assets(
            session,
            dispatch_id,
            res.lattice,
            res_record.lattice,
            local_store,
        )
        et = datetime.now()
        delta = (et - st).total_seconds()
        app_log.debug(f"{dispatch_id}: Inserting lattice took {delta} seconds")

        st = datetime.now()
        tg = import_transport_graph(
            session,
            dispatch_id,
            res.lattice.transport_graph,
            res_record.lattice,
            local_store,
            electron_id,
        )
        et = datetime.now()
        delta = (et - st).total_seconds()
        app_log.debug(f"{dispatch_id}: Inserting transport graph took {delta} seconds")

    lat = LatticeSchema(metadata=res.lattice.metadata, assets=lat_assets, transport_graph=tg)

    output = ResultSchema(metadata=res.metadata, assets=res_assets, lattice=lat)
    st = datetime.now()
    filtered_uris = _filter_remote_uris(output)
    et = datetime.now()
    delta = (et - st).total_seconds()
    app_log.debug(f"{dispatch_id}: Filtering URIs took {delta} seconds")
    return filtered_uris


def _connect_result_to_electron(
    session: Session, res: ResultSchema, parent_electron_id: int
) -> ResultSchema:
    """Link a sublattice dispatch to its parent electron"""

    # Update the `electron_id` lattice field and propagate the
    # `Job.cancel_requested` to the sublattice dispatch's jobs.

    app_log.debug("connecting previously submitted subdispatch to parent electron")
    sub_result = Result.from_dispatch_id(res.metadata.dispatch_id, bare=True)

    sub_result.set_value("electron_id", parent_electron_id, session)
    sub_result.set_value("root_dispatch_id", res.metadata.root_dispatch_id, session)

    parent_electron_record = ElectronMeta.get(
        session,
        fields={"id", "parent_lattice_id", "job_id"},
        equality_filters={"id": parent_electron_id},
        membership_filters={},
    )[0]
    parent_job_record = Job.get(
        session,
        fields={"id", "cancel_requested"},
        equality_filters={"id": parent_electron_record.job_id},
        membership_filters={},
        for_update=True,
    )[0]
    cancel_requested = parent_job_record.cancel_requested

    sub_electron_records = ElectronMeta.get(
        session,
        fields={"id", "parent_lattice_id", "job_id"},
        equality_filters={"parent_lattice_id": sub_result._lattice_id},
        membership_filters={},
    )

    job_ids = [rec.job_id for rec in sub_electron_records]

    Job.update_bulk(
        session,
        values={"cancel_requested": cancel_requested},
        equality_filters={},
        membership_filters={"id": job_ids},
    )

    return res


def _filter_remote_uris(manifest: ResultSchema) -> ResultSchema:
    dispatch_id = manifest.metadata.dispatch_id

    # Workflow-level
    for key, asset in manifest.assets:
        if asset.remote_uri:
            filtered_uri = filter_asset_uri(
                URI_FILTER_POLICY,
                asset.remote_uri,
                {},
                AssetScope.DISPATCH,
                dispatch_id,
                None,
                key,
            )
            asset.remote_uri = filtered_uri

    for key, asset in manifest.lattice.assets:
        if asset.remote_uri:
            filtered_uri = filter_asset_uri(
                URI_FILTER_POLICY, asset.remote_uri, {}, AssetScope.LATTICE, dispatch_id, None, key
            )
            asset.remote_uri = filtered_uri

    # Now filter each node
    tg = manifest.lattice.transport_graph
    for node in tg.nodes:
        for key, asset in node.assets:
            if asset.remote_uri:
                filtered_uri = filter_asset_uri(
                    URI_FILTER_POLICY,
                    asset.remote_uri,
                    {},
                    AssetScope.NODE,
                    dispatch_id,
                    node.id,
                    key,
                )
                asset.remote_uri = filtered_uri

    return manifest


def _get_result_meta(res: ResultSchema, storage_path: str, electron_id: Optional[int]) -> dict:
    kwargs = {
        "dispatch_id": res.metadata.dispatch_id,
        "root_dispatch_id": res.metadata.root_dispatch_id,
        "status": res.metadata.status,
        "started_at": res.metadata.start_time,
        "completed_at": res.metadata.end_time,
    }
    db_kwargs = {
        "electron_id": electron_id,
    }
    kwargs.update(db_kwargs)

    return kwargs


def import_result_assets(
    session: Session,
    manifest: ResultSchema,
    record: Result,
    object_store: BaseProvider,
) -> ResultAssets:
    """Insert asset records and populate the asset link table"""
    asset_ids = {}

    for asset_key, asset in manifest.assets:
        storage_path, object_key = object_store.get_uri_components(
            dispatch_id=manifest.metadata.dispatch_id,
            node_id=None,
            asset_key=asset_key,
        )
        local_uri = os.path.join(storage_path, object_key)

        asset_kwargs = {
            "storage_type": object_store.scheme,
            "storage_path": storage_path,
            "object_key": object_key,
            "digest_alg": asset.digest_alg,
            "digest": asset.digest,
            "remote_uri": asset.uri,
            "size": asset.size,
        }
        asset_ids[asset_key] = Asset.create(session, insert_kwargs=asset_kwargs, flush=False)

        # Send this back to the client
        asset.digest = None
        asset.remote_uri = f"file://{local_uri}"

    # Write asset records to DB
    n_records = len(asset_ids)

    st = datetime.now()
    session.flush()
    et = datetime.now()
    delta = (et - st).total_seconds()
    app_log.debug(f"Inserting {n_records} asset records took {delta} seconds")

    result_asset_links = [
        record.associate_asset(session, key, asset_rec.id) for key, asset_rec in asset_ids.items()
    ]
    n_records = len(result_asset_links)
    st = datetime.now()
    session.flush()
    et = datetime.now()
    delta = (et - st).total_seconds()
    app_log.debug(f"Inserting {n_records} asset links took {delta} seconds")

    return manifest.assets


# To be called after import_result
def handle_redispatch(
    manifest: ResultSchema,
    parent_dispatch_id: str,
    reuse_previous_results: bool,
) -> Tuple[ResultSchema, List[Tuple[Asset, Asset]]]:
    # * Compare transport graphs (tg_ops)
    # * Copy reusable nodes (tg_ops)
    # * Handle reuse_previous_results
    # * Filter node statuses in the DB: PENDING_REPLACEMENT -> NEW_OBJECT
    # * Filter asset upload URIs for reusable nodes
    # * Return filtered manifest

    dispatch_id = manifest.metadata.dispatch_id

    # Load the full NX graph for graph diffing (only node metadata
    # will actually be loaded in memory).
    result_object = Result.from_dispatch_id(dispatch_id, bare=False)
    parent_result_object = Result.from_dispatch_id(parent_dispatch_id, bare=False)

    tg_new = result_object.lattice.transport_graph
    tg_old = parent_result_object.lattice.transport_graph

    # Get the nodes that can potentially be reused from the previous
    # dispatch, assuming that they have previously completed.
    reusable_nodes = TransportGraphOps(tg_old).get_reusable_nodes(tg_new)

    # No need to upload assets for reusable nodes since they can be
    # copied internally from the previous dispatch. Thus, don't return
    # an upload URI to the client.
    reusable_nodes_set = set(reusable_nodes)
    tg_manifest = manifest.lattice.transport_graph

    with Result.session() as session:
        for node in tg_manifest.nodes:
            if node.id in reusable_nodes_set:
                dal_node = tg_new.get_node(node.id)
                for key, asset in node.assets:
                    asset.remote_uri = ""

                    # Don't pull asset
                    dal_asset = dal_node.get_asset(key, session)
                    dal_asset.set_remote(session, "")

    # Two cases:
    #
    # If not reuse_previous_results, copy assets for all reusable
    # nodes but leave all metadata as initialized by the SDK.  This
    # will cause all nodes to be rerun since their statuses will be
    # NEW_OBJECT.

    # If reuse_previous_results, copy all assets and metadata from the
    # previous dispatch. This will cause reusable nodes with a
    # COMPLETED corresponding node in the old dispatch to be marked
    # PENDING_REUSE in the DB, signalling to the dispatcher that they
    # don't need to be re-run.
    assets_to_copy = TransportGraphOps(tg_new).copy_nodes_from(
        tg_old,
        reusable_nodes,
        copy_metadata=reuse_previous_results,
        defer_copy_objects=True,
    )

    # Since the graph comparison is finished, we can upgrade
    # PENDING_REPLACEMENT to NEW_OBJECT in the DB.
    TransportGraphOps(tg_new).reset_nodes()

    # Copy corresponding workflow assets with the same hashes and
    # don't ask the client to upload them.

    with Result.session() as session:
        for key, asset in manifest.assets:
            new_asset = result_object.get_asset(key, session)
            old_asset = parent_result_object.get_asset(key, session)
            if new_asset.digest == old_asset.digest:
                asset.remote_uri = ""
                app_log.debug(f"Copying workflow asset {key}")
                assets_to_copy.append((old_asset, new_asset))
                # Don't pull asset
                new_asset.set_remote(session, "")

        for key, asset in manifest.lattice.assets:
            new_asset = result_object.lattice.get_asset(key, session)
            old_asset = parent_result_object.lattice.get_asset(key, session)
            if new_asset.digest == old_asset.digest:
                asset.remote_uri = ""
                app_log.debug(f"Copying workflow asset {key}")
                assets_to_copy.append((old_asset, new_asset))
                # Don't pull asset
                new_asset.set_remote(session, "")

    # Copy asset metadata
    with Result.session() as session:
        for item in assets_to_copy:
            src, dest = item
            copy_asset_meta(session, src, dest)

    # Copy asset data
    # for item in assets_to_copy:
    #     src, dest = item
    #     copy_asset(src, dest)

    return manifest, assets_to_copy



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DB Config"""

from sqlalchemy.ext.declarative import declarative_base

from covalent_dispatcher._db.datastore import DataStore

Base = declarative_base()
engine = DataStore.factory().engine


def init_db(db_path: str = None):
    global engine
    engine = (
        DataStore(db_URL=db_path, initialize_db=True).engine
        if db_path is not None
        else DataStore().engine
    )



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect
from typing import Any, Tuple

import cloudpickle
from pennylane._device import Device

from .logger import app_log
from .pickling import _qml_mods_pickle

_IMPORT_PATH_SEPARATOR = ":"


def get_qelectron_db_path(dispatch_id: str, task_id: int):
    """
    Return the path to the Qelectron database for a given dispatch_id and task_id.

    WARNING: SHOULD ONLY BE USED FROM THE SAME MACHINE
    AS WHERE THE USER'S TASK FUNCTION IS BEING RUN.
    """

    from covalent.quantum.qserver.database import Database

    database = Database()

    db_path = database.get_db_path(dispatch_id=dispatch_id, node_id=task_id)

    if db_path.exists():
        app_log.debug(f"Found qelectron DB for task {task_id}")
        return db_path
    else:
        app_log.debug(f"Qelectron database not found for task {task_id}")
        return None


@_qml_mods_pickle
def cloudpickle_serialize(obj):
    return cloudpickle.dumps(obj)


def cloudpickle_deserialize(obj):
    return cloudpickle.loads(obj)


def select_first_executor(qnode, executors):
    """Selects the first executor to run the qnode"""
    return executors[0]


def get_import_path(obj) -> Tuple[str, str]:
    """
    Determine the import path of an object.
    """
    if module := inspect.getmodule(obj):
        module_path = module.__name__
        class_name = obj.__name__
        return f"{module_path}{_IMPORT_PATH_SEPARATOR}{class_name}"
    raise RuntimeError(f"Unable to determine import path for {obj}.")


def import_from_path(path: str) -> Any:
    """
    Import a class from a path.
    """
    module_path, class_name = path.split(_IMPORT_PATH_SEPARATOR)
    module = importlib.import_module(module_path)
    return getattr(module, class_name)


def get_original_shots(dev: Device):
    """
    Recreate vector of shots if device has a shot vector.
    """
    if not dev.shot_vector:
        return dev.shots

    shot_sequence = []
    for shots in dev.shot_vector:
        shot_sequence.extend([shots.shots] * shots.copies)
    return type(dev.shot_vector)(shot_sequence)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mappings between electron attributes and DB records"""


from ..._db import models


def _to_endpoints(e_record: models.ElectronDependency, uid_node_id_map: dict):
    source = uid_node_id_map[e_record.parent_electron_id]
    target = uid_node_id_map[e_record.electron_id]
    return source, target


def _to_edge_attrs(e_record: models.ElectronDependency):
    attrs = {
        "edge_name": e_record.edge_name,
        "param_type": e_record.parameter_type,
        "arg_index": e_record.arg_index,
    }

    return attrs



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Asset class and utility functions"""

import os
from enum import Enum
from pathlib import Path
from typing import Any

from sqlalchemy.orm import Session

from covalent._shared_files import logger

from .._db.models import Asset as AssetRecord
from .._object_store.local import BaseProvider, local_store
from .controller import Record
from .utils.file_transfer import cp

app_log = logger.app_log


class StorageType(Enum):
    LOCAL = "file"
    S3 = "s3"


_storage_provider_map = {
    StorageType.LOCAL: local_store,
}


FIELDS = {
    "id",
    "storage_type",
    "storage_path",
    "object_key",
    "digest_alg",
    "digest",
    "remote_uri",
    "size",
}


class Asset(Record[AssetRecord]):

    """Metadata for an object in blob storage"""

    model = AssetRecord

    def __init__(self, session: Session, record: AssetRecord, *, keys: list = FIELDS):
        self._id = record.id
        self._attrs = {k: getattr(record, k) for k in keys}

    @property
    def object_store(self) -> BaseProvider:
        return _storage_provider_map[self.storage_type]

    @property
    def primary_key(self):
        return self._id

    @property
    def storage_type(self) -> StorageType:
        return StorageType(self._attrs["storage_type"])

    @property
    def storage_path(self) -> str:
        return self._attrs["storage_path"]

    @property
    def object_key(self) -> str:
        return self._attrs["object_key"]

    @property
    def digest_alg(self) -> str:
        return self._attrs["digest_alg"]

    @property
    def digest(self) -> str:
        return self._attrs["digest"]

    @property
    def remote_uri(self) -> str:
        return self._attrs["remote_uri"]

    @property
    def internal_uri(self) -> str:
        scheme = self.storage_type.value
        return f"{scheme}://" + str(Path(self.storage_path) / self.object_key)

    @property
    def size(self) -> int:
        return self._attrs["size"]

    def set_remote(self, session: Session, uri: str):
        self.update(session, values={"remote_uri": uri})

    def store_data(self, data: Any, session: Session) -> None:
        digest, size = self.object_store.store_file(self.storage_path, self.object_key, data)
        self.update(
            session,
            values={
                "digest_alg": digest.algorithm,
                "digest": digest.hexdigest,
                "size": size,
            },
        )

    def load_data(self) -> Any:
        return self.object_store.load_file(self.storage_path, self.object_key)

    def download(self, src_uri: str):
        scheme = self.storage_type.value
        dest_uri = scheme + "://" + os.path.join(self.storage_path, self.object_key)
        app_log.debug(f"Downloading asset from {src_uri} to {dest_uri}")

        cp(src_uri, dest_uri)

    def upload(self, dest_uri: str):
        scheme = self.storage_type.value
        src_uri = scheme + "://" + os.path.join(self.storage_path, self.object_key)
        app_log.debug(f"Uploading asset from {src_uri} to {dest_uri}")
        cp(src_uri, dest_uri)

    @classmethod
    def from_id(cls, asset_id: int, session: Session, *, keys=FIELDS) -> "Asset":
        records = cls.get(
            session, fields=keys, equality_filters={"id": asset_id}, membership_filters={}
        )
        record = records[0]
        return Asset(session, record, keys=keys)


def copy_asset(src: Asset, dest: Asset):
    """Copy the data for an asset.

    Args:
        session: SQLalchemy session
        src: The source asset
        dest The destination asset
    """

    if src.size > 0:
        scheme = dest.storage_type.value
        dest_uri = scheme + "://" + os.path.join(dest.storage_path, dest.object_key)
        src.upload(dest_uri)
    else:
        app_log.debug(f"Refusing to copy zero-sized asset {src.internal_uri}")


def copy_asset_meta(session: Session, src: Asset, dest: Asset):
    """Copy the metadata for an asset.

    Args:
        session: SQLalchemy session
        src: The source asset
        dest The destination asset
    """

    update = {
        "digest_alg": src.digest_alg,
        "digest": src.digest,
        "size": src.size,
    }
    dest.update(session, values=update)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to convert node -> ElectronSchema"""

from typing import Dict

from .._shared_files.schemas.electron import (
    ASSET_FILENAME_MAP,
    AssetSchema,
    ElectronAssets,
    ElectronMetadata,
    ElectronSchema,
)
from .._shared_files.util_classes import RESULT_STATUS, Status
from .common import AssetType, load_asset, save_asset

__all__ = [
    "serialize_node",
    "deserialize_node",
]


ASSET_TYPES = {
    "function": AssetType.TRANSPORTABLE,
    "function_string": AssetType.TEXT,
    "value": AssetType.TRANSPORTABLE,
    "output": AssetType.TRANSPORTABLE,
    "hooks": AssetType.JSONABLE,
    "qelectron_db": AssetType.BYTES,
    "stdout": AssetType.TEXT,
    "stderr": AssetType.TEXT,
    "error": AssetType.TEXT,
}


def _serialize_node_metadata(node_attrs: dict, node_storage_path: str) -> ElectronMetadata:
    task_group_id = node_attrs["task_group_id"]
    name = node_attrs["name"]
    executor = node_attrs["metadata"]["executor"]
    executor_data = node_attrs["metadata"]["executor_data"]
    qelectron_data_exists = node_attrs["metadata"]["qelectron_data_exists"]

    # Optional
    status = node_attrs.get("status", RESULT_STATUS.NEW_OBJECT)

    start_time = node_attrs.get("start_time")
    if start_time:
        start_time = start_time.isoformat()

    end_time = node_attrs.get("end_time")
    if end_time:
        end_time = end_time.isoformat()

    return ElectronMetadata(
        task_group_id=task_group_id,
        name=name,
        executor=executor,
        executor_data=executor_data,
        qelectron_data_exists=qelectron_data_exists,
        status=str(status),
        start_time=start_time,
        end_time=end_time,
    )


def _deserialize_node_metadata(meta: ElectronMetadata) -> dict:
    return {
        "task_group_id": meta.task_group_id,
        "name": meta.name,
        "status": Status(meta.status),
        "start_time": meta.start_time,
        "end_time": meta.end_time,
        "sub_dispatch_id": meta.sub_dispatch_id,
        "metadata": {
            "executor": meta.executor,
            "executor_data": meta.executor_data,
            "qelectron_data_exists": meta.qelectron_data_exists,
        },
    }


def _serialize_node_assets(node_attrs: dict, node_storage_path: str) -> ElectronAssets:
    function = node_attrs["function"]
    function_asset = save_asset(
        function,
        ASSET_TYPES["function"],
        node_storage_path,
        ASSET_FILENAME_MAP["function"],
    )

    function_string = node_attrs.get("function_string", None)
    function_string_asset = save_asset(
        function_string,
        ASSET_TYPES["function_string"],
        node_storage_path,
        ASSET_FILENAME_MAP["function_string"],
    )

    node_value = node_attrs.get("value", None)
    value_asset = save_asset(
        node_value,
        ASSET_TYPES["value"],
        node_storage_path,
        ASSET_FILENAME_MAP["value"],
    )

    node_output = node_attrs.get("output", None)
    output_asset = save_asset(
        node_output,
        ASSET_TYPES["output"],
        node_storage_path,
        ASSET_FILENAME_MAP["output"],
    )

    node_stdout = node_attrs.get("stdout", None)
    stdout_asset = save_asset(
        node_stdout,
        ASSET_TYPES["stdout"],
        node_storage_path,
        ASSET_FILENAME_MAP["stdout"],
    )

    node_stderr = node_attrs.get("stderr", None)
    stderr_asset = save_asset(
        node_stderr,
        ASSET_TYPES["stderr"],
        node_storage_path,
        ASSET_FILENAME_MAP["stderr"],
    )

    qelectron_db = node_attrs.get("qelectron_db", None)
    qelectron_db_asset = save_asset(
        qelectron_db,
        ASSET_TYPES["qelectron_db"],
        node_storage_path,
        ASSET_FILENAME_MAP["qelectron_db"],
    )

    node_error = node_attrs.get("error", None)
    error_asset = save_asset(
        node_error,
        ASSET_TYPES["error"],
        node_storage_path,
        ASSET_FILENAME_MAP["error"],
    )

    hooks = node_attrs["metadata"]["hooks"]
    hooks_asset = save_asset(
        hooks, ASSET_TYPES["hooks"], node_storage_path, ASSET_FILENAME_MAP["hooks"]
    )
    return ElectronAssets(
        function=function_asset,
        function_string=function_string_asset,
        value=value_asset,
        output=output_asset,
        stdout=stdout_asset,
        stderr=stderr_asset,
        qelectron_db=qelectron_db_asset,
        error=error_asset,
        hooks=hooks_asset,
    )


def _deserialize_node_assets(ea: ElectronAssets) -> dict:
    function = load_asset(ea.function, ASSET_TYPES["function"])
    function_string = load_asset(ea.function_string, ASSET_TYPES["function_string"])
    value = load_asset(ea.value, ASSET_TYPES["value"])
    output = load_asset(ea.output, ASSET_TYPES["output"])
    stdout = load_asset(ea.stdout, ASSET_TYPES["stdout"])
    stderr = load_asset(ea.stderr, ASSET_TYPES["stderr"])
    qelectron_db = load_asset(ea.qelectron_db, ASSET_TYPES["qelectron_db"])
    error = load_asset(ea.error, ASSET_TYPES["error"])

    hooks = load_asset(ea.hooks, ASSET_TYPES["hooks"])

    return {
        "function": function,
        "function_string": function_string,
        "value": value,
        "output": output,
        "stdout": stdout,
        "stderr": stderr,
        "qelectron_db": qelectron_db,
        "error": error,
        "metadata": {
            "hooks": hooks,
        },
    }


def _get_node_custom_assets(node_attrs: dict) -> Dict[str, AssetSchema]:
    if "custom_asset_keys" in node_attrs["metadata"]:
        return {key: AssetSchema(size=0) for key in node_attrs["metadata"]["custom_asset_keys"]}


def serialize_node(node_id: int, node_attrs: dict, node_storage_path) -> ElectronSchema:
    meta = _serialize_node_metadata(node_attrs, node_storage_path)
    assets = _serialize_node_assets(node_attrs, node_storage_path)
    custom_assets = _get_node_custom_assets(node_attrs)
    return ElectronSchema(id=node_id, metadata=meta, assets=assets, custom_assets=custom_assets)


def deserialize_node(e: ElectronSchema, metadata_only: bool = False) -> dict:
    node_attrs = _deserialize_node_metadata(e.metadata)
    node_assets = _deserialize_node_assets(e.assets)

    asset_metadata = node_assets.pop("metadata")

    if not metadata_only:
        node_attrs.update(node_assets)

    # merge "metadata" attrs
    node_attrs["metadata"].update(asset_metadata)

    return {"id": e.id, "attrs": node_attrs}



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines executors and provides a "manager" to get all available executors
"""

import contextlib
import glob
import importlib
import inspect
import os
import sys
from pathlib import Path
from typing import Any, Dict, List, Union

import pkg_resources

from .._shared_files import logger
from .._shared_files.config import get_config, update_config
from .base import BaseExecutor

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_QUANTUM_PLUGINS_PATH = Path(__file__).parent / "quantum_plugins"
_QUANTUM_DEFAULTS_VARNAME = "_QEXECUTOR_PLUGIN_DEFAULTS"


class _ExecutorManager:
    """
    Executor manager to return a valid executor which can be
    used as an argument to electron and lattice decorators.

    Initializing generates a list of available executor plugins.
    """

    def __init__(self) -> None:
        # Dictionary mapping executor name to executor class
        self.executor_plugins_map: Dict[str, Any] = {}
        self.executor_plugins_exports_map: Dict[str, Any] = {}

        if os.environ.get("COVALENT_PLUGIN_LOAD", "true").lower() == "true":
            self.generate_plugins_list()

    def __new__(cls):
        # Singleton pattern for this class
        if not hasattr(cls, "instance"):
            cls.instance = super().__new__(cls)
        return cls.instance

    def generate_plugins_list(self) -> None:
        """
        Generate a list of available executor plugins.
        This is called automatically when the class is initialized.

        The list of executors is generated by loading the already
        installed plugins and the plugins in the executor directory.

        These executor plugins are loaded by importing the module that
        contains the plugin.

        The module should have an attribute named executor_plugin_name
        which is set to the class name defining the plugin.

        Args:
            None

        Returns:
            None
        """

        # Load plugins that are part of the covalent path:
        pkg_plugins_path = os.path.join(os.path.dirname(__file__), "executor_plugins")
        self._load_executors(pkg_plugins_path)

        # Look for executor plugins in a user-defined path:
        user_plugins_path = ":".join(
            filter(
                None,
                [
                    get_config("sdk.executor_dir"),
                    os.environ.get("COVALENT_EXECUTOR_DIR"),
                ],
            )
        )
        self._load_executors(user_plugins_path)

        # Look for pip-installed plugins:
        self._load_installed_plugins()

    def get_executor(self, name: Union[str, BaseExecutor]) -> BaseExecutor:
        """
        Get an executor by name.
        This accepts a string like "local" or a BaseExecutor instance.

        Args:
            name: The name of the executor to get.

        Returns:
            executor: An executor object.

        Raises:
            ValueError: If name is not found and new is False.
            TypeError: If name is not a string or a BaseExecutor instance.
        """

        if isinstance(name, BaseExecutor):
            return name

        elif isinstance(name, str):
            if name in self.executor_plugins_map:
                update_config()
                default_options = get_config(f"executors.{name}")
                return self.executor_plugins_map[name](**default_options)
            else:
                message = f"No executor found by name: {name}."
                app_log.error(message)
                raise ValueError(f"No executor found by name: {name}")
        else:
            message = f"Input must be of type str or BaseExecutor, not {type(name)}"
            app_log.error(message)
            raise TypeError

    def _is_plugin_name_valid(self, the_module):
        """Assert if the plugin variable name is valid"""

        plugin_name_var = "EXECUTOR_PLUGIN_NAME"
        return bool(
            hasattr(the_module, plugin_name_var) or hasattr(the_module, plugin_name_var.lower())
        )

    def nonzero_plugin_classes(self, plugin_class):
        """Retrun true if any plugin classes are present"""

        return bool(len(plugin_class))

    def _populate_executor_map_from_module(self, the_module: Any) -> None:
        """
        Populate the executor map from a module.
        Also checks whether `EXECUTOR_PLUGIN_NAME` is defined in the module.

        Args:
            the_module: The module to populate the executor map from.

        Returns:
            None
        """

        if not self._is_plugin_name_valid(the_module):
            message = f"{the_module.__name__} does not seem to have a well-defined plugin class.\n"
            message += f"Specify the plugin class with 'EXECUTOR_PLUGIN_NAME = <plugin class name>' in the {the_module.__name__} module."
            app_log.warning(message)
            return

        # All classes loaded by the module
        all_classes = inspect.getmembers(the_module, inspect.isclass)
        # Classes that are defined in the module:
        module_classes = [c[1] for c in all_classes if c[1].__module__ == the_module.__name__]
        # The module should have a global attribute named EXECUTOR_PLUGIN_NAME
        # which is set to the class name defining the plugin.
        executor_name = (
            the_module.executor_plugin_name
            if hasattr(the_module, "executor_plugin_name")
            else the_module.EXECUTOR_PLUGIN_NAME
        )
        plugin_class = [c for c in module_classes if c.__name__ == executor_name]

        if self.nonzero_plugin_classes(plugin_class):
            plugin_class = plugin_class[0]
            short_name = the_module.__name__.split("/")[-1].split(".")[-1]
            self.executor_plugins_map[short_name] = plugin_class

            if hasattr(the_module, "_EXECUTOR_PLUGIN_DEFAULTS"):
                default_params = {
                    "executors": {short_name: getattr(the_module, "_EXECUTOR_PLUGIN_DEFAULTS")},
                }
                update_config(default_params, override_existing=False)

        else:
            # The requested plugin (the_module.module_name) was not found in the module.
            executor_name = (
                the_module.executor_plugin_name
                if hasattr(the_module, "executor_plugin_name")
                else the_module.EXECUTOR_PLUGIN_NAME
            )
            message = (
                f"Requested executor plugin {executor_name} was not found in {the_module.__name__}"
            )
            app_log.warning(message)

    def _load_installed_plugins(self) -> None:
        """
        Load executor plugins from pip.
        Populates the executor_plugins_map attribute.

        Args:
            None

        Returns:
            None
        """

        entry_points = pkg_resources.iter_entry_points("covalent.executor.executor_plugins")
        for entry in entry_points:
            the_module = entry.load()
            self._populate_executor_map_from_module(the_module)

    def _load_executors(self, executor_dir: str) -> None:
        """
        Load executor plugins from a directory.
        Populates the executor map and executor_plugins_map attributes with user's plugins.

        Args:
            executor_dir: The directory to load executor plugins from.

        Returns:
            None
        """

        dirs = set(executor_dir.split(":"))

        for e_dir in dirs:
            if os.path.exists(e_dir):
                module_files = glob.glob(os.path.join(e_dir, "*.py"))

                for module_file in module_files:
                    if module_file.endswith("__init__.py"):
                        continue

                    module_name = module_file[:-3]

                    # Import the module that contains the plugin
                    module_spec = importlib.util.spec_from_file_location(module_name, module_file)
                    the_module = importlib.util.module_from_spec(module_spec)
                    module_spec.loader.exec_module(the_module)

                    self._populate_executor_map_from_module(the_module)

    def list_executors(self, regenerate: bool = False, print_names: bool = True) -> List[str]:
        """
        Return and optionally print the executors that are available.

        Args:
            regenerate: If True, the executor map is regenerated.
            print_names: If True, executor names are printed as well.

        Returns:
            A list of executor names.
        """

        if regenerate:
            self.generate_plugins_list()

        executor_list = []
        for n, name in enumerate(self.executor_plugins_map, start=1):
            executor_list.append(name)
            if print_names:
                print(f"{n}. {name}")

        return executor_list


class _QExecutorManager:
    """
    QExecutor manager to return a valid QExecutor which can be
    used as an argument to `qelectron` decorator.

    Initializing generates a list of available QExecutor plugins.
    """

    def __init__(self):
        from ..quantum import QCluster, Simulator

        # Dictionary mapping executor name to executor class
        self.executor_plugins_map: Dict[str, Any] = {
            "QCluster": QCluster,
            "Simulator": Simulator,
        }
        self.load_executors()

    def __new__(cls):
        # Singleton pattern for this class
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance

    def load_executors(self) -> None:
        """
        Looks for `plugin.py` modules in the subdirectories of the given path and
        loads QExecutor classes from them.
        """

        # Iterate over all subdirectories of the plugins path except for those starting with "_" like "__pycache__"
        for plugin_dir in filter(
            lambda _p: _p.is_dir() and not _p.name.startswith("_"), _QUANTUM_PLUGINS_PATH.iterdir()
        ):
            # Get the Path of the plugin module
            plugin_module_path = list(plugin_dir.glob("*_plugin.py"))
            if not plugin_module_path:
                continue
            plugin_module_path = plugin_module_path[0]

            if plugin_module_path.exists():
                # Suppress any exceptions that may occur while importing the plugin module
                # This is to prevent the plugin module from crashing the application
                with contextlib.suppress(Exception):
                    sys.path.append(str(plugin_module_path.parent))
                    plugin_module_spec = importlib.util.spec_from_file_location(
                        plugin_module_path.stem, plugin_module_path
                    )
                    plugin_module = importlib.util.module_from_spec(plugin_module_spec)
                    sys.modules[plugin_module_path.stem] = plugin_module
                    plugin_module_spec.loader.exec_module(plugin_module)
                    self.populate_executors_map(plugin_module)

        # print("Loaded QExecutor plugins: ", self.executor_plugins_map.keys())

    def populate_executors_map(self, module_obj) -> None:
        """
        Validates modules containing potential QExecutor classes.
        Populates the `executor_plugins_map` dictionary and updates the config.
        """

        self.validate_module(module_obj)

        for qexecutor_cls_name in getattr(module_obj, "__all__"):
            self.executor_plugins_map[qexecutor_cls_name] = getattr(module_obj, qexecutor_cls_name)

        for qexecutor_cls_name, defaults_dict in getattr(
            module_obj, _QUANTUM_DEFAULTS_VARNAME
        ).items():
            update_config(
                {"qelectron": {qexecutor_cls_name: defaults_dict}}, override_existing=False
            )

    def validate_module(self, module_obj) -> None:
        """
        Checks all of the following:
            - module exports plugin classes using `__all__`
            - module defines `_QUANTUM_DEFAULTS_VARNAME`
            - exported plugin classes match the default parameters

        """
        if not hasattr(module_obj, "__all__"):
            return
        if not hasattr(module_obj, _QUANTUM_DEFAULTS_VARNAME):
            return

        plugin_defaults = getattr(module_obj, _QUANTUM_DEFAULTS_VARNAME)

        if set(module_obj.__all__).difference(set(plugin_defaults)):
            raise RuntimeError(f"Module missing default parameters in {module_obj.__file__}")
        if set(plugin_defaults).difference(set(module_obj.__all__)):
            raise RuntimeError(
                f"Non-exported QExecutor class in default parameters in {module_obj.__file__}"
            )


_executor_manager = _ExecutorManager()
for name in _executor_manager.executor_plugins_map:
    plugin_class = _executor_manager.executor_plugins_map[name]
    globals()[plugin_class.__name__] = plugin_class

# Only creating the qexecutor manager if its requirements are installed
with contextlib.suppress(ImportError):
    _qexecutor_manager = _QExecutorManager()
    for qexecutor_cls in _qexecutor_manager.executor_plugins_map.values():
        globals()[qexecutor_cls.__name__] = qexecutor_cls



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Endpoints for dispatch management"""

import asyncio
import json
from contextlib import asynccontextmanager
from typing import List, Optional, Union
from uuid import UUID

from fastapi import APIRouter, FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse

import covalent_dispatcher.entry_point as dispatcher
from covalent._shared_files import logger
from covalent._shared_files.schemas.result import ResultSchema
from covalent._shared_files.util_classes import RESULT_STATUS
from covalent_dispatcher._core import dispatcher as core_dispatcher
from covalent_dispatcher._core import runner_ng as core_runner

from .._dal.exporters.result import export_result_manifest
from .._dal.result import Result, get_result_object
from .._db.datastore import workflow_db
from .._db.dispatchdb import DispatchDB
from .heartbeat import Heartbeat
from .models import DispatchStatusSetSchema, ExportResponseSchema, TargetDispatchStatus

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router: APIRouter = APIRouter()

_background_tasks = set()


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Initialize global variables"""

    heartbeat = Heartbeat()
    fut = asyncio.create_task(heartbeat.start())
    _background_tasks.add(fut)
    fut.add_done_callback(_background_tasks.discard)

    # Runner event queue and listener
    core_runner._job_events = asyncio.Queue()
    core_runner._job_event_listener = asyncio.create_task(core_runner._listen_for_job_events())

    # Dispatcher event queue and listener
    core_dispatcher._global_status_queue = asyncio.Queue()
    core_dispatcher._global_event_listener = asyncio.create_task(
        core_dispatcher._node_event_listener()
    )

    yield

    # Cancel all scheduled and running dispatches
    for status in [
        RESULT_STATUS.NEW_OBJECT,
        RESULT_STATUS.RUNNING,
    ]:
        await cancel_all_with_status(status)

    core_dispatcher._global_event_listener.cancel()
    core_runner._job_event_listener.cancel()

    Heartbeat.stop()


async def cancel_all_with_status(status: RESULT_STATUS):
    """Cancel all dispatches with the specified status."""

    with workflow_db.session() as session:
        records = Result.get_db_records(
            session,
            keys=["dispatch_id"],
            equality_filters={"status": str(status)},
            membership_filters={},
        )

        for record in records:
            dispatch_id = record.dispatch_id
            await dispatcher.cancel_running_dispatch(dispatch_id)


@router.post("/dispatches/submit")
async def submit(request: Request) -> UUID:
    """
    Function to accept the submit request of
    new dispatch and return the dispatch id
    back to the client.

    Args:
        None

    Returns:
        dispatch_id: The dispatch id in a json format
                     returned as a Fast API Response object.
    """
    try:
        data = await request.json()
        data = json.dumps(data).encode("utf-8")
        return await dispatcher.make_dispatch(data)
    except Exception as e:
        raise HTTPException(
            status_code=400,
            detail=f"Failed to submit workflow: {e}",
        ) from e


async def start(dispatch_id: str):
    """Start a previously registered (re-)dispatch.

    Args:
        `dispatch_id`: The dispatch's unique id.

    Returns:
        `dispatch_id`
    """
    fut = asyncio.create_task(dispatcher.start_dispatch(dispatch_id))
    _background_tasks.add(fut)
    fut.add_done_callback(_background_tasks.discard)

    return dispatch_id


async def cancel(dispatch_id: str, task_ids: List[int] = None) -> str:
    """
    Function to handle the cancel request of
    a dispatch.

    Args:
        dispatch_id: ID of the dispatch
        task_ids: (Query) Optional list of specific task ids to cancel.
            An empty list will cause all tasks to be cancelled.

    Returns:
        Fast API Response object confirming that the dispatch
        has been cancelled.
    """

    if task_ids is None:
        task_ids = []

    await dispatcher.cancel_running_dispatch(dispatch_id, task_ids)
    if task_ids:
        return f"Cancelled tasks {task_ids} in dispatch {dispatch_id}."
    else:
        return f"Dispatch {dispatch_id} cancelled."


@router.get("/db-path")
def db_path() -> str:
    db_path = DispatchDB()._dbpath
    return json.dumps(db_path)


@router.post("/dispatches", status_code=201)
async def register(manifest: ResultSchema) -> ResultSchema:
    """Register a dispatch in the database.

    Args:
        manifest: Declares all metadata and assets in the workflow
        parent_dispatch_id: The parent dispatch id if registering a sublattice dispatch

    Returns:
        The manifest with `dispatch_id` and remote URIs for each asset populated.
    """
    try:
        return await dispatcher.register_dispatch(manifest, None)
    except Exception as e:
        app_log.debug(f"Exception in register: {e}")
        raise HTTPException(
            status_code=400,
            detail=f"Failed to submit workflow: {e}",
        ) from e


@router.post("/dispatches/{dispatch_id}/subdispatches", status_code=201)
async def register_subdispatch(
    manifest: ResultSchema,
    dispatch_id: str,
) -> ResultSchema:
    """Register a subdispatch in the database.

    Args:
        manifest: Declares all metadata and assets in the workflow
        dispatch_id: The parent dispatch id

    Returns:
        The manifest with `dispatch_id` and remote URIs for each asset populated.
    """
    try:
        return await dispatcher.register_dispatch(manifest, dispatch_id)
    except Exception as e:
        app_log.debug(f"Exception in register: {e}")
        raise HTTPException(
            status_code=400,
            detail=f"Failed to submit workflow: {e}",
        ) from e


@router.post("/dispatches/{dispatch_id}/redispatches", status_code=201)
async def register_redispatch(
    manifest: ResultSchema,
    dispatch_id: str,
    reuse_previous_results: bool = False,
):
    """Register a redispatch in the database.

    Args:
        manifest: Declares all metadata and assets in the workflow
        dispatch_id: The original dispatch's id.
        reuse_previous_results: Whether to try reusing the results of
            previously completed electrons.

    Returns:
        The manifest with `dispatch_id` and remote URIs for each asset populated.
    """
    try:
        return await dispatcher.register_redispatch(
            manifest,
            dispatch_id,
            reuse_previous_results,
        )
    except Exception as e:
        app_log.debug(f"Exception in register_redispatch: {e}")
        raise HTTPException(
            status_code=400,
            detail=f"Failed to submit workflow: {e}",
        ) from e


@router.put("/dispatches/{dispatch_id}/status", status_code=202)
async def set_dispatch_status(dispatch_id: str, desired_status: DispatchStatusSetSchema):
    """Set the status of a dispatch.

    Valid target statuses are:
        - "RUNNING" to start a dispatch
        - "CANCELLED" to cancel dispatch processing

    Args:
        `dispatch_id`: The dispatch's unique id
        `desired_status`: A `StatusSetSchema` object describing the desired status.

    """

    if desired_status.status == TargetDispatchStatus.running:
        return await start(dispatch_id)
    else:
        return await cancel(dispatch_id, desired_status.task_ids)


@router.get("/dispatches/{dispatch_id}")
async def export_result(
    dispatch_id: str, wait: Optional[bool] = False, status_only: Optional[bool] = False
) -> ExportResponseSchema:
    """Export all metadata about a registered dispatch

    Args:
        `dispatch_id`: The dispatch's unique id.

    Returns:
        {
            id: `dispatch_id`,
            status: status,
            result_export: manifest for the result
        }

    The manifest `result_export` has the same schema as that which is
    submitted to `/register`.

    """
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(
        None,
        _export_result_sync,
        dispatch_id,
        wait,
        status_only,
    )


def _export_result_sync(
    dispatch_id: str, wait: Optional[bool] = False, status_only: Optional[bool] = False
) -> ExportResponseSchema:
    result_object = _try_get_result_object(dispatch_id)
    if not result_object:
        return JSONResponse(
            status_code=404,
            content={"message": f"The requested dispatch ID {dispatch_id} was not found."},
        )
    status = str(result_object.get_value("status", refresh=False))

    if not wait or status in [
        str(RESULT_STATUS.COMPLETED),
        str(RESULT_STATUS.FAILED),
        str(RESULT_STATUS.CANCELLED),
    ]:
        output = {
            "id": dispatch_id,
            "status": status,
        }
        if not status_only:
            output["result_export"] = export_result_manifest(dispatch_id)

        return output

    response = JSONResponse(
        status_code=503,
        content={"message": "Result not ready to read yet. Please wait for a couple of seconds."},
        headers={"Retry-After": "2"},
    )
    return response


def _try_get_result_object(dispatch_id: str) -> Union[Result, None]:
    try:
        res = get_result_object(
            dispatch_id, bare=True, keys=["id", "dispatch_id", "status"], lattice_keys=["id"]
        )
    except KeyError:
        res = None
    return res



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from .file import File


class Folder(File):
    """
    Folder class to store components of provided URI including scheme (s3://, file://, ect.), determine if the file is remote,
    and act as facade to facilitate filesystem operations. Folder is a child of the File class which sets `is_dir` flag to True.

    Attributes:
        include_folder: Flag that determines if the folder should be included in the file transfer, if False only contents of folder are transfered.
    """

    def __init__(
        self,
        filepath: Optional[str] = None,
        is_remote: bool = False,
        is_dir: bool = True,
        include_folder: bool = False,
    ):
        super().__init__(filepath, is_remote, is_dir, include_folder)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DB-backed transport graph"""

from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Tuple

import networkx as nx
from sqlalchemy.orm import Session

from covalent._shared_files import logger

from .._db.models import ElectronDependency as EdgeRecord
from .db_interfaces.tg_utils import (
    _all_edge_records,
    _child_records,
    _edge_records_for_nodes,
    _incoming_edge_records,
)
from .edge import Edge
from .electron import ELECTRON_KEYS
from .electron import Electron as Node

app_log = logger.app_log


class _TransportGraph:
    def __init__(self, lattice_id: int, bare: bool = False, *, keys: List = ELECTRON_KEYS):
        self.lattice_id = lattice_id
        self.bare = bare
        self._nodes = {}
        self._graph = nx.MultiDiGraph()
        self._keys = keys

    def add_node(self, node: Node):
        self._graph.add_node(node.node_id, **node.metadata.attrs)
        self._nodes[node.node_id] = node

    def add_edge(self, x: int, y: int, **attrs):
        self._graph.add_edge(x, y, **attrs)

    def get_node(self, node_id: int, session: Session = None) -> Node:
        return self.get_nodes(node_ids=[node_id], session=session)[0]

    def get_nodes(self, node_ids: List[int], session: Session = None) -> List[Node]:
        if not self.bare:
            return [self._nodes[node_id] for node_id in node_ids]

        # Construct node from db
        start = datetime.now()
        if session:
            nodes = _nodes(session, self.lattice_id, node_ids, keys=self._keys)
        else:
            with Node.session() as session:
                nodes = _nodes(session, self.lattice_id, node_ids, keys=self._keys)

        end = datetime.now()
        dt = (end - start).total_seconds()
        app_log.debug(f"get_nodes {node_ids} took {dt} seconds")
        return nodes

    def get_node_value(
        self, node_id: int, key: str, session: Session = None, refresh: bool = True
    ):
        records = self.get_values_for_nodes(
            node_ids=[node_id], keys=[key], session=session, refresh=refresh
        )
        return records[0][key]

    def get_values_for_nodes(
        self, node_ids: List[int], keys: List[str], session: Session = None, refresh: bool = True
    ) -> List[Dict]:
        """Bulk query attributes for nodes.

        Args:
            node_ids: The list of nodes to query
            keys: The list of attributes to query
            session: An optional SQLalchemy session to use for the DB query
            refresh: A boolean indicating whether to use cached attributes

        If session is `None`, a temporary session will be created for
        the query.

        Returns:
            A list of dictionaries {attr_key: attr_val}, one for
            each node id, in the same order as `node_ids`.

        """

        nodes = self.get_nodes(node_ids=node_ids, session=session)
        return list(map(lambda n: n.get_values(keys, session, refresh), nodes))

    def set_node_value(self, node_id: int, key: str, val: Any, session: Session = None):
        node = self.get_node(node_id, session)
        node.set_value(key, val, session)

    def get_incoming_edges(self, node_id: int) -> List[Tuple[int, int, Dict]]:
        """Query in-edges of a node.

        Returns:
            List[Edge], where

            Edge is a dictionary with structure
                source: int,
                target: int,
                attrs: dict
        """

        # Read from internal NX graph
        if not self.bare:
            pred = list(self._graph.predecessors(node_id))
            edge_list = [
                {"source": s, "target": node_id, "attrs": d}
                for s in pred
                for _, d in self.get_edge_data(s, node_id).items()
            ]
            return edge_list

        # Read from DB
        with Node.session() as session:
            node = self.get_node(node_id, session)
            edge_list = _get_incoming_edges(session, node, keys=self._keys)
            return list(
                map(
                    lambda e: {"source": e.source, "target": e.target, "attrs": e.attrs}, edge_list
                )
            )

    def get_successors(self, node_id: int, attr_keys: List = []) -> List[Dict]:
        """Get child nodes with multiplicity.

        Parameters:
            node_id: id of node
            attr_keys: list of node attributes to return, such as task_group_id

        Returns:
            List[Dict], where each dictionary is of the form
            {"node_id": node_id, attr_key_1: node_attr[attr_key_1], ...}

        """

        # Read from internal NX graph
        if not self.bare:
            node_list = [
                self.get_node(child)
                for child, edges in self._graph.adj[node_id].items()
                for edge in edges
            ]
            return _filter_node_list(node_list, None, attr_keys)

        # Query DB
        with Node.session() as session:
            node = self.get_node(node_id, session)
            child_node_list = _get_child_nodes(session, node, keys=attr_keys)
            return _filter_node_list(child_node_list, session, attr_keys)

    # Copied from _TransportGraph
    def get_edge_data(self, dep_key: int, node_key: int) -> Any:
        """
        Get the metadata for all edges between two nodes.

        Args:
            dep_key: The node id for first node.
            node_key: The node id for second node.

        Returns:
            values: A dict {edge_key : value}

        Raises:
            KeyError: If the edge is not found.
        """

        if not self.bare:
            return self._graph.get_edge_data(dep_key, node_key)

        with Node.session() as session:
            source = self.get_node(dep_key, session)
            target = self.get_node(node_key, session)
            return _get_edge_data_for_nodes(session, source, target)

    def get_internal_graph_copy(self) -> nx.MultiDiGraph:
        return self._graph.copy()

    def get_dependencies(self, node_key: int) -> list:
        """Gets the parent node ids of a node with multiplicity

        Args:
            node_key: The node id.

        Returns: parents: The dependencies of the node. Parent nodes
            are repeated according to edge multiplicity.

        """
        return [e["source"] for e in self.get_incoming_edges(node_key)]

    @staticmethod
    def get_compute_graph(
        session: Session, lattice_id: int, bare: bool = False, *, keys: List = ELECTRON_KEYS
    ) -> _TransportGraph:
        if not bare:
            nodes, edges = _nodes_and_edges(session, lattice_id, keys=keys)
            return _make_compute_graph(lattice_id, nodes, edges, keys=keys)
        else:
            app_log.debug("Getting bare transport graph")
            return _TransportGraph(lattice_id, True, keys=keys)


def _get_incoming_edges(session: Session, node: Node, *, keys: List) -> List[Edge]:
    records = _incoming_edge_records(session, node._electron_id, keys=keys)
    nodes = list(map(lambda r: Node(session, r[0], keys=keys), records))
    uid_node_id_map = {n._electron_id: n.node_id for n in nodes}
    uid_node_id_map[node._electron_id] = node.node_id
    edge_list = list(map(lambda r: _to_edge(r[1], uid_node_id_map), records))

    return edge_list


def _get_child_nodes(session: Session, node: Node, *, keys: List) -> List[Node]:
    """Return successor nodes with multiplicity"""
    records = _child_records(session, node._electron_id, keys=keys)
    return list(map(lambda r: Node(session, r, keys=keys), records))


def _to_edge(e_record: EdgeRecord, uid_node_id_map: Dict) -> Edge:
    return Edge(e_record, uid_node_id_map)


def _nodes(session: Session, lattice_id: int, node_ids: List[int], *, keys: List) -> List[Node]:
    # records = _node_records(session, lattice_id, node_ids)
    records = Node.get_db_records(
        session,
        keys=keys,
        equality_filters={"parent_lattice_id": lattice_id},
        membership_filters={"node_id": node_ids},
    )
    if len(records) < len(node_ids):
        raise KeyError(f"Invalid Node ids {node_ids} for lattice record {lattice_id}")
    return list(map(lambda x: Node(session, x, keys=keys), records))


def _get_edge_data_for_nodes(session: Session, parent_node: Node, child_node: Node):
    records = _edge_records_for_nodes(session, parent_node._electron_id, child_node._electron_id)

    uid_node_id_map = {
        child_node._electron_id: child_node.node_id,
        parent_node._electron_id: parent_node.node_id,
    }
    edge_list = list(map(lambda r: _to_edge(r, uid_node_id_map), records))

    return {i: e.attrs for i, e in enumerate(edge_list)}


def _nodes_and_edges(
    session: Session, lattice_id: int, *, keys: List
) -> Tuple[List[Node], List[Edge]]:
    db_nodes = Node.get_db_records(
        session,
        keys=keys,
        equality_filters={"parent_lattice_id": lattice_id},
        membership_filters={},
    )
    db_edges = _all_edge_records(session, lattice_id)
    uid_nodeid_map = {e.id: e.transport_graph_node_id for e in db_nodes}
    nodes = list(map(lambda x: Node(session, x, keys=keys), db_nodes))
    edges = list(map(lambda x: _to_edge(x, uid_nodeid_map), db_edges))

    return nodes, edges


def _make_compute_graph(
    lattice_id: int, nodes: List, edges: List, *, keys: List
) -> _TransportGraph:
    tg = _TransportGraph(lattice_id, keys=keys)
    for node in nodes:
        tg.add_node(node)
    for edge in edges:
        tg.add_edge(edge.source, edge.target, **edge.attrs)
    return tg


def _filter_node(node_obj: Node, session: Session, attr_keys: List[str]):
    output = {"node_id": node_obj.node_id}
    for key in attr_keys:
        output[key] = node_obj.get_value(key, session, refresh=False)
    return output


def _filter_node_list(node_list: List[Node], session: Session, attr_keys: List[str]):
    return list(map(lambda x: _filter_node(x, session, attr_keys), node_list))



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for querying the transport graph
"""

from typing import Dict, List

from ..._dal.result import get_result_object
from .utils import run_in_executor


def get_bulk_sync(dispatch_id: str, node_ids: List[int], keys: List[str]) -> List[Dict]:
    result_object = get_result_object(dispatch_id)
    attrs = result_object.lattice.transport_graph.get_values_for_nodes(
        node_ids=node_ids,
        keys=keys,
        refresh=False,
    )
    return attrs


async def get_bulk(dispatch_id: str, node_ids: List[int], keys: List[str]) -> List[Dict]:
    """Query attributes for multiple electrons.

    Args:
        node_ids: The list of nodes to query
        keys: The list of attributes to query for each electron

    Returns:
        A list of dictionaries {attr_key: attr_val}, one for
        each node id, in the same order as `node_ids`

    Example:
    ```
        await get_bulk(
            "my_dispatch", [2, 4], ["name", "status"],
        )
    ```
    will return
    ```
    [
        {
            "name": "task_2", "status": RESULT_STATUS.COMPLETED,
        },
        {
            "name": "task_4, "status": RESULT_STATUS.FAILED,
        },
    ]
    ```

    """
    return await run_in_executor(
        get_bulk_sync,
        dispatch_id,
        node_ids,
        keys,
    )


async def get(dispatch_id: str, node_id: int, keys: List[str]) -> Dict:
    """Convenience function to query attributes for an electron.

    Args:
        node_id: The node to query
        keys: The list of attributes to query

    Returns:
        A dictionary {attr_key: attr_val}

    Example:
    ```
        await get(
            "my_dispatch", 2, ["name", "status"],
        )
    ```
    will return
    ```
        {
            "name": "task_2", "status": RESULT_STATUS.COMPLETED,
        }
    ```

    """
    attrs = await get_bulk(dispatch_id, [node_id], keys)
    return attrs[0]


def update_sync(dispatch_id: str, node_result: Dict):
    result_object = get_result_object(dispatch_id, bare=True)
    return result_object._update_node(**node_result)


async def update(dispatch_id: str, node_result: Dict):
    """Update a node's attributes"""
    return await run_in_executor(update_sync, dispatch_id, node_result)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from covalent.triggers import BaseTrigger, DatabaseTrigger, DirTrigger, SQLiteTrigger, TimeTrigger


class TriggerLoader:
    def __init__(self):
        self.available_triggers = {
            BaseTrigger.__name__: BaseTrigger,
            DatabaseTrigger.__name__: DatabaseTrigger,
            DirTrigger.__name__: DirTrigger,
            TimeTrigger.__name__: TimeTrigger,
            SQLiteTrigger.__name__: SQLiteTrigger,
        }

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super().__new__(cls)
        return cls.instance

    def __getitem__(self, key):
        return self.available_triggers.get(key)

    def __setitem__(self, key, value):
        self.available_triggers[key] = value


available_triggers = TriggerLoader()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines the core functionality of the runner
"""

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent.executor import _executor_manager
from covalent.executor.base import AsyncBaseExecutor

app_log = logger.app_log
log_stack_info = logger.log_stack_info
debug_mode = get_config("sdk.log_level") == "debug"


def get_executor(node_id, selected_executor, loop=None, pool=None) -> AsyncBaseExecutor:
    # Instantiate the executor from JSON

    short_name, object_dict = selected_executor

    app_log.debug(f"Running task {node_id} using executor {short_name}, {object_dict}")

    # the executor is determined during scheduling and provided in the execution metadata
    executor = _executor_manager.get_executor(short_name)
    executor.from_dict(object_dict)
    executor._init_runtime(loop=loop, cancel_pool=pool)

    return executor



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mappings between graph attributes and DB records"""


from typing import List

from sqlalchemy import select
from sqlalchemy.orm import Load, Session

from ..._db.models import Electron as ElectronRecord
from ..._db.models import ElectronDependency as EdgeRecord
from ..._db.models import Lattice as LatticeRecord
from .. import electron


def _edge_records_for_nodes(
    session: Session, parent_electron_id: int, child_electron_id: int
) -> List[EdgeRecord]:
    stmt = (
        select(EdgeRecord)
        .where(EdgeRecord.electron_id == child_electron_id)
        .where(EdgeRecord.parent_electron_id == parent_electron_id)
    )
    records = session.scalars(stmt).all()
    if not records:
        raise KeyError(f"No edges between nodes {parent_electron_id}, {child_electron_id}")
    return list(records)


def _incoming_edge_records(
    session: Session, electron_id: int, *, keys: List
) -> List[ElectronRecord]:
    stmt = (
        select(ElectronRecord, EdgeRecord)
        .join(EdgeRecord, EdgeRecord.parent_electron_id == ElectronRecord.id)
        .where(EdgeRecord.electron_id == electron_id)
    )
    if len(keys) > 0:
        fields = list(map(electron.Electron.meta_record_map, keys))
        attrs = [getattr(ElectronRecord, f) for f in fields]
        stmt = stmt.options(Load(ElectronRecord).load_only(*attrs))

    records = session.execute(stmt).all()
    return list(map(lambda r: (r.Electron, r.ElectronDependency), records))


def _child_records(session: Session, electron_id: int, *, keys: List) -> List[ElectronRecord]:
    stmt = (
        select(ElectronRecord)
        .join(EdgeRecord, EdgeRecord.electron_id == ElectronRecord.id)
        .where(EdgeRecord.parent_electron_id == electron_id)
    )
    if len(keys) > 0:
        fields = list(map(electron.Electron.meta_record_map, keys))
        attrs = [getattr(ElectronRecord, f) for f in fields]
        stmt = stmt.options(Load(ElectronRecord).load_only(*attrs))

    records = session.execute(stmt).all()
    return list(map(lambda r: r.Electron, records))


# Join electron dependency with filtered electrons on destination electron
def _all_edge_records(session: Session, lattice_id: int) -> List[EdgeRecord]:
    stmt = (
        select(EdgeRecord)
        .join(ElectronRecord, ElectronRecord.id == EdgeRecord.electron_id)
        .join(LatticeRecord, LatticeRecord.id == ElectronRecord.parent_lattice_id)
        .where(LatticeRecord.id == lattice_id)
    )
    records = session.execute(stmt).all()
    return list(map(lambda r: r.ElectronDependency, records))



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains all the functions required interface with the jobs table"""

from typing import Dict, List

from sqlalchemy import select
from sqlalchemy.orm import Session

from covalent._shared_files import logger

from .datastore import workflow_db
from .models import Electron, Job, Lattice

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class MissingJobRecordError(Exception):
    """
    Exception to be raised when the job record associated with a particular task is not found in the database
    """

    def __init__(self, message: str = ""):
        self.message = message
        super().__init__(self.message)


def transaction_get_job_record(session: Session, job_id: int) -> Dict:
    """
    Query the database for the job record associated with the given job id using the passed in session
    Session object is passed to ensure that for all db related calls the same session is used

    Arg(s)
        session: SQLalchemy session object
        job_id: ID of the job to query

    Return(s)
        Dictionary of the job record from the database
    """
    if job_record := session.query(Job).where(Job.id == job_id).first():
        return {
            "job_id": job_record.id,
            "cancel_requested": job_record.cancel_requested,
            "status": job_record.status,
            "job_handle": job_record.job_handle,
        }
    else:
        raise MissingJobRecordError(message=f"Job {job_id} not found")


def _update_job_record(
    session: Session,
    job_id: int,
    cancel_requested: bool = None,
    job_handle: str = None,
    job_status: str = None,
):
    """
    Update the job record in the database

    Arg(s)
        session: SQLalchemy session object
        job_id: ID of the job to update
        cancel_requested: Boolean flag indicating whether the job was requested to be cancelled
        cancel_successful: Boolean indicating whether the job was cancelled successfully
        job_handle: Unique job handle returned by the execution backend

    Return(s)
        None
    """
    job_record = session.query(Job).where(Job.id == job_id).first()
    if not job_record:
        raise MissingJobRecordError(message=f"Job {job_id} not found")

    if cancel_requested is not None:
        job_record.cancel_requested = cancel_requested
    if job_handle is not None:
        job_record.job_handle = job_handle
    if job_status is not None:
        job_record.status = job_status


def get_job_record(job_id: int) -> Dict:
    """
    Retrive the job record from database

    Arg(s)
        job_id: ID of the job to be returned

    Return(s)
        Job record of the job identified by `job_id`
    """
    return get_job_records([job_id])[0]


def update_job_records(record_kwargs_list: list):
    """
    Update job records in the database

    Arg(s)
        record_kwargs_list: List of keyword arguments of the fields that need to be updated in the job records

    Return(s)
        None
    """
    with workflow_db.session() as session:
        for entry in record_kwargs_list:
            _update_job_record(session, **entry)


def get_job_records(job_ids: List[int]) -> List[Dict]:
    """
    Retrieve the job records of all jobs with `job_ids`

    Arg(s)
        job_ids: List of job ids to query the job records of

    Return(s)
        Job records of all tasks with `job_ids`
    """
    with workflow_db.session() as session:
        records = list(map(lambda x: transaction_get_job_record(session, x), job_ids))
    return records


def to_job_ids(dispatch_id: str, task_ids: List[int]) -> List[int]:
    """
    Map all lattice task ids to their corresponding job ids

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_ids: IDs of tasks in the lattice

    Return(s)
        Corresponding job ids assocated with the provided task ids
    """
    with workflow_db.session() as session:
        stmt = select(Lattice).where(Lattice.dispatch_id == dispatch_id)
        lattice_rec = session.scalars(stmt).first()
        if not lattice_rec:
            raise KeyError(f"Invalid dispatch {dispatch_id}")

        stmt = (
            select(Electron.job_id)
            .where(Electron.parent_lattice_id == lattice_rec.id)
            .where(Electron.transport_graph_node_id.in_(task_ids))
        )

        records = session.scalars(stmt).all()

        return records
